# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" A grid control with a model/ui architecture. """


import sys
import wx
import wx.lib.gridmovers as grid_movers
from os.path import abspath, exists
from wx.grid import Grid as wxGrid
from wx.grid import GridCellAttr, GridCellBoolRenderer, GridTableBase
from wx.grid import (
    GridTableMessage,
    GRIDTABLE_NOTIFY_ROWS_APPENDED,
    GRIDTABLE_NOTIFY_ROWS_DELETED,
    GRIDTABLE_NOTIFY_ROWS_INSERTED,
    GRIDTABLE_NOTIFY_COLS_APPENDED,
    GRIDTABLE_NOTIFY_COLS_DELETED,
    GRIDTABLE_NOTIFY_COLS_INSERTED,
    GRIDTABLE_REQUEST_VIEW_GET_VALUES,
    GRID_VALUE_STRING,
)
from wx import TheClipboard


from pyface.api import Widget
from pyface.timer.api import do_later
from traits.api import (
    Bool,
    Color,
    Enum,
    Event,
    Font,
    Instance,
    Int,
    Trait,
    Undefined,
)
from pyface.wx.drag_and_drop import (
    PythonDropSource,
    PythonDropTarget,
    PythonObject,
)
from pyface.wx.drag_and_drop import clipboard as enClipboard, FileDropSource

from .grid_model import GridModel

# Is this code running on MS Windows?
is_win32 = sys.platform == "win32"

ASCII_C = 67


class Grid(Widget):
    """ A grid control with a model/ui architecture. """

    # 'Grid' interface -----------------------------------------------------

    # The model that provides the data for the grid.
    model = Instance(GridModel, ())

    # Should grid lines be shown on the table?
    enable_lines = Bool(True)

    # The color to show gridlines in
    grid_line_color = Color("blue")

    # Show row headers?
    show_row_headers = Bool(True)

    # Show column headers?
    show_column_headers = Bool(True)

    # The default font to use for text in labels
    default_label_font = Font(None)

    # The default background color for labels
    default_label_bg_color = Color(wx.Colour(236, 233, 216))

    # The default text color for labels
    default_label_text_color = Color("black")

    # The color to use for a selection background
    selection_bg_color = Trait(wx.Colour(49, 106, 197), None, Color)

    # The color to use for a selection foreground/text
    selection_text_color = Trait(wx.Colour(255, 255, 255), None, Color)

    # The default font to use for text in cells
    default_cell_font = Font(None)

    # The default text color to use for text in cells
    default_cell_text_color = Color("black")

    # The default background color to use for editable cells
    default_cell_bg_color = Color("white")

    # The default background color to use for read-only cells
    # default_cell_read_only_color = Trait(Color("linen"), None, Color)
    default_cell_read_only_color = Color(wx.Colour(248, 247, 241))

    # Should the grid be read-only? If this is set to false, individual
    # cells can still declare themselves read-only.
    read_only = Bool(False)

    # Selection mode.
    selection_mode = Enum("cell", "rows", "cols", "")

    # Sort data when a column header is clicked?
    allow_column_sort = Bool(True)

    # Sort data when a row header is clicked?
    allow_row_sort = Bool(False)

    # pixel height of column labels
    column_label_height = Int(32)

    # pixel width of row labels
    row_label_width = Int(82)

    # auto-size columns and rows?
    autosize = Bool(False)

    # Allow single-click access to cell-editors?
    edit_on_first_click = Bool(True)

    # Events ----

    # A cell has been activated (ie. double-clicked).
    cell_activated = Event()

    # The current selection has changed.
    selection_changed = Event()

    # A drag operation was started on a cell.
    cell_begin_drag = Event()

    # A left-click occurred on a cell.
    cell_left_clicked = Event()

    # A left-click occurred on a cell at specific location
    # Useful if the cell contains multiple controls though the hit test
    # is left to the consumer of the event
    cell_left_clicked_location = Event()

    # A right-click occurred on a cell.
    cell_right_clicked = Event()

    # protected variables to store the location of the clicked event
    _x_clicked = Int()
    _y_clicked = Int()

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------
    def __init__(self, parent, **traits):
        """ Creates a new grid.

        'parent' is the toolkit-specific control that is the grid's parent.

        """

        # Base class constructors.
        super(Grid, self).__init__(**traits)

        # Flag set when columns are resizing:
        self._user_col_size = False

        # Create the toolkit-specific control.
        self.control = self._grid = grid = wxGrid(parent, -1)
        grid.grid = self

        self._moveTo = None
        self._edit = False
        grid.Bind(wx.EVT_IDLE, self._on_idle)
        # Set when moving edit cursor:
        grid._no_reset_col = False
        grid._no_reset_row = False

        # initialize the current selection
        self.__current_selection = ()

        self.__initialize_counts(self.model)
        self.__initialize_sort_state()

        # Don't display any extra space around the rows and columns.
        grid.SetMargins(0, 0)

        # Provides more accurate scrolling behavior without creating large
        # margins on the bottom and right. The down side is that it makes
        # scrolling using the scroll bar buttons painfully slow:
        grid.SetScrollLineX(1)
        grid.SetScrollLineY(1)

        # Tell the grid to get its data from the model.
        #
        # N.B The terminology used in the wxPython API is a little confusing!
        # --- The 'SetTable' method is actually setting the model used by
        #     the grid (which is the view)!
        #
        # The second parameter to 'SetTable' tells the grid to take ownership
        # of the model and to destroy it when it is done.  Otherwise you would
        # need to keep a reference to the model and manually destroy it later
        # (by calling it's Destroy method).
        #
        # fixme: We should create a default model if one is not supplied.

        # The wx virtual table hook.

        self._grid_table_base = _GridTableBase(self.model, self)

        # keep the onership of the table in this class
        grid.SetTable(self._grid_table_base, takeOwnership=False)

        # Enable column and row moving:
        grid_movers.GridColMover(grid)
        grid_movers.GridRowMover(grid)
        grid.Bind(grid_movers.EVT_GRID_COL_MOVE, self._on_col_move)
        grid.Bind(grid_movers.EVT_GRID_ROW_MOVE, self._on_row_move)

        self.model.observe(self._on_model_content_changed, "content_changed")
        self.model.observe(
            self._on_model_structure_changed, "structure_changed"
        )
        self.model.observe(self._on_row_sort, "row_sorted")
        self.model.observe(self._on_column_sort, "column_sorted")
        self.observe(self._on_new_model, "model")

        # hook up style trait handlers - note that we have to use
        # dynamic notification hook-ups because these handlers should
        # not be called until after the control object is initialized.
        # static trait notifiers get called when the object inits.
        self.observe(self._on_enable_lines_changed, "enable_lines")
        self.observe(self._on_grid_line_color_changed, "grid_line_color")
        self.observe(self._on_default_label_font_changed, "default_label_font")
        self.observe(
            self._on_default_label_bg_color_changed, "default_label_bg_color"
        )
        self.observe(
            self._on_default_label_text_color_changed,
            "default_label_text_color",
        )
        self.observe(self._on_selection_bg_color_changed, "selection_bg_color")
        self.observe(
            self._on_selection_text_color_changed, "selection_text_color"
        )
        self.observe(self._on_default_cell_font_changed, "default_cell_font")
        self.observe(
            self._on_default_cell_text_color_changed, "default_cell_text_color"
        )
        self.observe(
            self._on_default_cell_bg_color_changed, "default_cell_bg_color"
        )
        self.observe(self._on_read_only_changed, "read_only_changed")
        self.observe(self._on_selection_mode_changed, "selection_mode")
        self.observe(
            self._on_column_label_height_changed, "column_label_height"
        )
        self.observe(self._on_row_label_width_changed, "row_label_width")
        self.observe(
            self._on_show_column_headers_changed, "show_column_headers"
        )
        self.observe(self._on_show_row_headers_changed, "show_row_headers")

        # Initialize wx handlers:
        self._notify_select = True
        grid.Bind(wx.grid.EVT_GRID_SELECT_CELL, self._on_select_cell)
        grid.Bind(wx.grid.EVT_GRID_RANGE_SELECT, self._on_range_select)
        grid.Bind(wx.grid.EVT_GRID_COL_SIZE, self._on_col_size)
        grid.Bind(wx.grid.EVT_GRID_ROW_SIZE, self._on_row_size)

        # This starts the cell editor on a double-click as well as on a second
        # click:
        grid.Bind(wx.grid.EVT_GRID_CELL_LEFT_DCLICK, self._on_cell_left_dclick)

        # Notify when cells are clicked on:
        grid.Bind(wx.grid.EVT_GRID_CELL_RIGHT_CLICK, self._on_cell_right_click)
        grid.Bind(
            wx.grid.EVT_GRID_CELL_RIGHT_DCLICK, self._on_cell_right_dclick
        )

        grid.Bind(
            wx.grid.EVT_GRID_LABEL_RIGHT_CLICK, self._on_label_right_click
        )
        grid.Bind(wx.grid.EVT_GRID_LABEL_LEFT_CLICK, self._on_label_left_click)

        if is_win32:
            grid.Bind(wx.grid.EVT_GRID_EDITOR_HIDDEN, self._on_editor_hidden)

        # We handle key presses to change the behavior of the <Enter> and
        # <Tab> keys to make manual data entry smoother.
        grid.Bind(wx.EVT_KEY_DOWN, self._on_key_down)

        # We handle control resize events to adjust column widths
        grid.Bind(wx.EVT_SIZE, self._on_size)

        # Handle drags:
        self._corner_window = grid.GetGridCornerLabelWindow()
        self._grid_window = gw = grid.GetGridWindow()
        self._row_window = rw = grid.GetGridRowLabelWindow()
        self._col_window = cw = grid.GetGridColLabelWindow()

        # Handle mouse button state changes:
        self._ignore = False
        for window in (gw, rw, cw):
            window.Bind(wx.EVT_MOTION, self._on_grid_motion)
            window.Bind(wx.EVT_LEFT_DOWN, self._on_left_down)
            window.Bind(wx.EVT_LEFT_UP, self._on_left_up)

        # Initialize the row and column models:
        self.__initialize_rows(self.model)
        self.__initialize_columns(self.model)
        self.__initialize_fonts()

        # Handle trait style settings:
        self.__initialize_style_settings()

        # Enable the grid as a drag and drop target:
        self._grid.SetDropTarget(PythonDropTarget(self))

        self.__autosize()

        self._edit = False
        grid.Bind(wx.EVT_IDLE, self._on_idle)

    def dispose(self):
        # Remove all wx handlers:
        grid = self._grid
        if grid is not None:
            grid.Unbind(wx.grid.EVT_GRID_SELECT_CELL)
            grid.Unbind(wx.grid.EVT_GRID_RANGE_SELECT)
            grid.Unbind(wx.grid.EVT_GRID_COL_SIZE)
            grid.Unbind(wx.grid.EVT_GRID_ROW_SIZE)
            grid.Unbind(wx.grid.EVT_GRID_CELL_LEFT_DCLICK)
            grid.Unbind(wx.grid.EVT_GRID_CELL_RIGHT_CLICK)
            grid.Unbind(wx.grid.EVT_GRID_CELL_RIGHT_DCLICK)
            grid.Unbind(wx.grid.EVT_GRID_LABEL_RIGHT_CLICK)
            grid.Unbind(wx.grid.EVT_GRID_LABEL_LEFT_CLICK)
            grid.Unbind(wx.grid.EVT_GRID_EDITOR_CREATED)
            if is_win32:
                grid.Unbind(wx.grid.EVT_GRID_EDITOR_HIDDEN)
            grid.Unbind(wx.EVT_KEY_DOWN)
            grid.Unbind(wx.EVT_SIZE)

        self._grid_window.Unbind(wx.EVT_PAINT)

        for window in (self._grid_window, self._row_window, self._col_window):
            window.Unbind(wx.EVT_MOTION)
            window.Unbind(wx.EVT_LEFT_DOWN)
            window.Unbind(wx.EVT_LEFT_UP)

        self.model.observe(
            self._on_model_content_changed, "content_changed", remove=True
        )
        self.model.observe(
            self._on_model_structure_changed, "structure_changed", remove=True
        )
        self.model.observe(self._on_row_sort, "row_sorted", remove=True)
        self.model.observe(self._on_column_sort, "column_sorted", remove=True)
        self.observe(self._on_new_model, "model", remove=True)

        self.observe(
            self._on_enable_lines_changed, "enable_lines", remove=True
        )
        self.observe(
            self._on_grid_line_color_changed, "grid_line_color", remove=True
        )
        self.observe(
            self._on_default_label_font_changed,
            "default_label_font",
            remove=True,
        )
        self.observe(
            self._on_default_label_bg_color_changed,
            "default_label_bg_color",
            remove=True,
        )
        self.observe(
            self._on_default_label_text_color_changed,
            "default_label_text_color",
            remove=True,
        )
        self.observe(
            self._on_selection_bg_color_changed,
            "selection_bg_color",
            remove=True,
        )
        self.observe(
            self._on_selection_text_color_changed,
            "selection_text_color",
            remove=True,
        )
        self.observe(
            self._on_default_cell_font_changed,
            "default_cell_font",
            remove=True,
        )
        self.observe(
            self._on_default_cell_text_color_changed,
            "default_cell_text_color",
            remove=True,
        )
        self.observe(
            self._on_default_cell_bg_color_changed,
            "default_cell_bg_color",
            remove=True,
        )
        self.observe(
            self._on_read_only_changed, "read_only_changed", remove=True
        )
        self.observe(
            self._on_selection_mode_changed, "selection_mode", remove=True
        )
        self.observe(
            self._on_column_label_height_changed,
            "column_label_height",
            remove=True,
        )
        self.observe(
            self._on_row_label_width_changed, "row_label_width", remove=True
        )
        self.observe(
            self._on_show_column_headers_changed,
            "show_column_headers",
            remove=True,
        )
        self.observe(
            self._on_show_row_headers_changed, "show_row_headers", remove=True
        )

        self._grid_table_base.dispose()
        self._grid = None

    # ------------------------------------------------------------------------
    # Trait event handlers.
    # ------------------------------------------------------------------------

    def _on_new_model(self, event):
        """ When we get a new model reinitialize grid match to that model. """

        self._grid_table_base.model = self.model

        self.__initialize_counts(self.model)

        self._on_model_changed()

        if self.autosize:
            # Note that we don't call AutoSize() here, because autosizing
            # the rows looks like crap.
            self._grid.AutoSizeColumns(False)

    def _on_model_content_changed(self, event):
        """ A notification method called when the data in the underlying
            model changes. """
        self._grid.ForceRefresh()

    def _on_model_structure_changed(self, event=None):
        """ A notification method called when the underlying model has
        changed. Responsible for making sure the view object updates
        correctly. """

        # Disable any active editors in order to prevent a wx crash bug:
        self._edit = False
        grid = self._grid

        # Make sure any current active editor has been disabled:
        grid.DisableCellEditControl()

        # More wacky fun with wx. We have to manufacture the appropriate
        # grid messages and send them off to make sure the grid updates
        # correctly:
        should_autosize = False

        # First check to see if rows have been added or deleted:
        row_count = self.model.get_row_count()
        delta = row_count - self._row_count
        self._row_count = row_count

        if delta > 0:
            # Rows were added:
            msg = GridTableMessage(
                self._grid_table_base, GRIDTABLE_NOTIFY_ROWS_APPENDED, delta
            )
            grid.ProcessTableMessage(msg)
            should_autosize = True
        elif delta < 0:
            # Rows were deleted:
            msg = GridTableMessage(
                self._grid_table_base,
                GRIDTABLE_NOTIFY_ROWS_DELETED,
                row_count,
                -delta,
            )
            grid.ProcessTableMessage(msg)
            should_autosize = True

        # Now check for column changes:
        col_count = self.model.get_column_count()
        delta = col_count - self._col_count
        self._col_count = col_count

        if delta > 0:
            # Columns were added:
            msg = GridTableMessage(
                self._grid_table_base, GRIDTABLE_NOTIFY_COLS_APPENDED, delta
            )
            grid.ProcessTableMessage(msg)
            should_autosize = True
        elif delta < 0:
            # Columns were deleted:
            msg = GridTableMessage(
                self._grid_table_base,
                GRIDTABLE_NOTIFY_COLS_DELETED,
                col_count,
                -delta,
            )
            grid.ProcessTableMessage(msg)
            should_autosize = True

        # Finally make sure we update for any new values in the table:
        msg = GridTableMessage(
            self._grid_table_base, GRIDTABLE_REQUEST_VIEW_GET_VALUES
        )
        grid.ProcessTableMessage(msg)

        if should_autosize:
            self.__autosize()

        # We have to make sure the editor/renderer cache in the GridTableBase
        # object is cleaned out:
        self._grid_table_base._clear_cache()

        grid.AdjustScrollbars()
        self._refresh()

    def _on_row_sort(self, event):
        """ Handles a row_sorted event from the underlying model. """

        # First grab the new data out of the event:
        if event.new.index < 0:
            self._current_sorted_row = None
        else:
            self._current_sorted_row = event.new.index

        self._row_sort_reversed = event.new.reversed

        # Since the label may have changed we may need to autosize again:
        # fixme: when we change how we represent the sorted column
        #        this should go away.
        self.__autosize()

        # Make sure everything updates to reflect the changes:
        self._on_model_structure_changed()

    def _on_column_sort(self, event):
        """ Handles a column_sorted event from the underlying model. """

        # first grab the new data out of the event
        if event.new.index < 0:
            self._current_sorted_col = None
        else:
            self._current_sorted_col = event.new.index

        self._col_sort_reversed = event.new.reversed

        # since the label may have changed we may need to autosize again
        # fixme: when we change how we represent the sorted column
        #        this should go away.
        self.__autosize()

        # make sure everything updates to reflect the changes
        self._on_model_structure_changed()

    def _on_enable_lines_changed(self, event=None):
        """ Handle a change to the enable_lines trait. """
        self._grid.EnableGridLines(self.enable_lines)

    def _on_grid_line_color_changed(self, event=None):
        """ Handle a change to the enable_lines trait. """
        self._grid.SetGridLineColour(self.grid_line_color)

    def _on_default_label_font_changed(self, event=None):
        """ Handle a change to the default_label_font trait. """

        font = self.default_label_font
        if font is None:
            font = wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
            font.SetWeight(wx.BOLD)

        self._grid.SetLabelFont(font)

    def _on_default_label_text_color_changed(self, event=None):
        """ Handle a change to the default_cell_text_color trait. """

        if self.default_label_text_color is not None:
            color = self.default_label_text_color
            self._grid.SetLabelTextColour(color)
            self._grid.ForceRefresh()

    def _on_default_label_bg_color_changed(self, event=None):
        """ Handle a change to the default_cell_text_color trait. """

        if self.default_label_bg_color is not None:
            color = self.default_label_bg_color
            self._grid.SetLabelBackgroundColour(color)
            self._grid.ForceRefresh()

    def _on_selection_bg_color_changed(self, event=None):
        """ Handle a change to the selection_bg_color trait. """
        if self.selection_bg_color is not None:
            self._grid.SetSelectionBackground(self.selection_bg_color)

    def _on_selection_text_color_changed(self, event=None):
        """ Handle a change to the selection_text_color trait. """
        if self.selection_text_color is not None:
            self._grid.SetSelectionForeground(self.selection_text_color)

    def _on_default_cell_font_changed(self, event=None):
        """ Handle a change to the default_cell_font trait. """

        if self.default_cell_font is not None:
            self._grid.SetDefaultCellFont(self.default_cell_font)
            self._grid.ForceRefresh()

    def _on_default_cell_text_color_changed(self, event=None):
        """ Handle a change to the default_cell_text_color trait. """

        if self.default_cell_text_color is not None:
            color = self.default_cell_text_color
            self._grid.SetDefaultCellTextColour(color)
            self._grid.ForceRefresh()

    def _on_default_cell_bg_color_changed(self, event=None):
        """ Handle a change to the default_cell_bg_color trait. """

        if self.default_cell_bg_color is not None:
            color = self.default_cell_bg_color
            self._grid.SetDefaultCellBackgroundColour(color)
            self._grid.ForceRefresh()

    def _on_read_only_changed(self, event=None):
        """ Handle a change to the read_only trait. """

        # should the whole grid be read-only?
        if self.read_only:
            self._grid.EnableEditing(False)
        else:
            self._grid.EnableEditing(True)

    def _on_selection_mode_changed(self, event=None):
        """ Handle a change to the selection_mode trait. """

        # should we allow individual cells to be selected or only rows
        # or only columns
        if self.selection_mode == "cell":
            self._grid.SetSelectionMode(wxGrid.SelectCells)
        elif self.selection_mode == "rows":
            self._grid.SetSelectionMode(wxGrid.SelectRows)
        elif self.selection_mode == "cols":
            self._grid.SetSelectionMode(wxGrid.SelectColumns)

    def _on_column_label_height_changed(self, event=None):
        """ Handle a change to the column_label_height trait. """

        # handle setting for height of column labels
        if self.column_label_height is not None:
            self._grid.SetColLabelSize(self.column_label_height)

    def _on_row_label_width_changed(self, event=None):
        """ Handle a change to the row_label_width trait. """
        # handle setting for width of row labels
        if self.row_label_width is not None:
            self._grid.SetRowLabelSize(self.row_label_width)

    def _on_show_column_headers_changed(self, event=None):
        """ Handle a change to the show_column_headers trait. """

        if not self.show_column_headers:
            self._grid.SetColLabelSize(0)
        else:
            self._grid.SetColLabelSize(self.column_label_height)

    def _on_show_row_headers_changed(self, event=None):
        """ Handle a change to the show_row_headers trait. """

        if not self.show_row_headers:
            self._grid.SetRowLabelSize(0)
        else:
            self._grid.SetRowLabelSize(self.row_label_width)

    # ------------------------------------------------------------------------
    # 'Grid' interface.
    # ------------------------------------------------------------------------

    def get_selection(self):
        """ Return a list of the currently selected objects. """

        return self.__get_selection()

    def set_selection(self, selection_list, notify=True):
        """ Set the current selection to the objects listed in selection_list.
        Note that these objects are model-specific, as the grid depends on the
        underlying model to translate these objects into grid coordinates.
        A ValueError will be raised if the objects are not in the proper format
        for the underlying model. """

        # Set the 'should we notify the model of the selection change' flag:
        self._notify_select = notify

        # First use the model to resolve the object list into a set of
        # grid coordinates
        cells = self.model.resolve_selection(selection_list)

        # Now make sure all those grid coordinates get set properly:
        grid = self._grid
        grid.BeginBatch()

        grid.ClearSelection()

        mode = self.selection_mode
        if mode == "rows":
            self._select_rows(cells)
        elif mode != "":
            for selection in cells:
                row, col = max(0, selection[0]), max(0, selection[1])
                grid.SelectBlock(row, col, row, col, True)

        grid.EndBatch()

        # Indicate that the selection has been changed:
        if notify:
            self.__fire_selection_changed()

        self._notify_select = True

    def stop_editing_indices(self, indices):
        """ If editing is occuring in a row in 'indices', stop editing. """

        if self._grid.GetGridCursorRow() in indices:
            self._grid.DisableCellEditControl()

    # ------------------------------------------------------------------------
    # wx event handlers.
    # ------------------------------------------------------------------------

    def _on_size(self, evt):
        """ Called when the grid is resized. """
        self.__autosize()
        evt.Skip()

    def _on_editor_hidden(self, evt):
        # This is a hack to make clicking on a window button while a grid
        # editor is active work correctly under Windows. Normally, when the
        # user clicks on the button to exit grid cell editing and perform the
        # button function, only the grid cell editing is terminated under
        # Windows. A second click on the button is required to actually
        # trigger the button functionality. We circumvent this problem by
        # generating a 'fake' button click event. Technically this solution
        # is not correct since the button click should be generated on mouse
        # up, but I'm not sure if we will get the mouse up event, so we do it
        # here instead. Note that this handler is only set-up when the OS is
        # Windows.
        control = wx.FindWindowAtPointer()
        if isinstance(control, wx.Button):
            do_later(
                wx.PostEvent,
                control,
                wx.CommandEvent(
                    wx.wxEVT_COMMAND_BUTTON_CLICKED, control.GetId()
                ),
            )

    def _on_left_down(self, evt):
        """ Called when the left mouse button is pressed.
        """
        grid = self._grid
        self._x_clicked = evt.GetX()
        self._y_clicked = evt.GetY()
        self._ignore = (grid.XToEdgeOfCol(evt.GetX()) != wx.NOT_FOUND) or (
            grid.YToEdgeOfRow(evt.GetY()) != wx.NOT_FOUND
        )
        evt.Skip()

    def _on_left_up(self, evt):
        """ Called when the left mouse button is released.
        """
        self._ignore = False
        evt.Skip()

    def _on_motion(self, evt):
        """ Called when the mouse moves. """

        evt.Skip()
        if evt.Dragging() and not evt.ControlDown():
            data = self.__get_drag_value()
            if isinstance(data, str):
                file = abspath(data)
                if exists(file):
                    FileDropSource(self._grid, file)
                    return

            PythonDropSource(self._grid, data)

    def _on_grid_motion(self, evt):
        if evt.GetEventObject() is self._grid_window:
            x, y = self._grid.CalcUnscrolledPosition(evt.GetPosition().Get())
            row = self._grid.YToRow(y)
            col = self._grid.XToCol(x)

            # Notify the model that the mouse has moved into the cell at row,col,
            # only if the row and col are valid.
            if (row >= 0) and (col >= 0):
                self.model.mouse_cell = (row, col)

            # If we are not ignoring mouse events, call _on_motion.
            if not self._ignore:
                self._on_motion(evt)

        evt.Skip()

    def _on_select_cell(self, evt):
        """ Called when the user has moved to another cell. """
        row, col = evt.GetRow(), evt.GetCol()
        self._moveTo = (row, col)
        self.cell_left_clicked = self.model.click = (row, col)

        # Try to find a renderer for this cell:
        renderer = self.model.get_cell_renderer(row, col)

        # If the renderer has defined a handler then call it:
        result = False
        if renderer is not None:
            result = renderer.on_left_click(self, row, col)

        # print("self._grid.GetParent()", self._grid.GetParent().GetParent().GetParent())
        # if the handler didn't tell us to stop further handling then skip
        if not result:
            if (self.selection_mode != "") or (not self.edit_on_first_click):
                self._grid.SelectBlock(row, col, row, col, evt.ControlDown())

            self._edit = True
            evt.Skip()

    def _on_range_select(self, evt):
        if evt.Selecting():
            if (self.selection_mode == "cell") and evt.ControlDown():
                self._grid.SelectBlock(
                    evt.GetTopRow(),
                    evt.GetLeftCol(),
                    evt.GetBottomRow(),
                    evt.GetRightCol(),
                    True,
                )

        if self._notify_select:
            self.__fire_selection_changed()

    def _on_col_size(self, evt):
        """ Called when the user changes a column's width. """
        self.__autosize()

        evt.Skip()

    def _on_row_size(self, evt):
        """ Called when the user changes a row's height. """
        self._grid.AdjustScrollbars()

        evt.Skip()

    def _on_idle(self, evt):
        """ Immediately jumps into editing mode, bypassing the
            usual select mode of a spreadsheet. See also self.OnSelectCell().
        """
        if self._edit == True and self.edit_on_first_click:
            if self._grid.CanEnableCellControl():
                self._grid.EnableCellEditControl()
            self._edit = False

        evt.Skip()

    def _on_cell_left_dclick(self, evt):
        """ Called when the left mouse button was double-clicked.

        From the wxPython demo code:-

        'I do this because I don't like the default behaviour of not starting
        the cell editor on double clicks, but only a second click.'

        Fair enuff!

        """
        row, col = evt.GetRow(), evt.GetCol()
        data = self.model.get_value(row, col)
        self.cell_activated = data

        # Tell the model that a cell was double-clicked on:
        self.model.dclick = (row, col)

        # Try to find a renderer for this cell
        renderer = self.model.get_cell_renderer(row, col)

        # If the renderer has defined a handler then call it
        if renderer is not None:
            renderer.on_left_dclick(self, row, col)

    def _on_cell_right_dclick(self, evt):
        """ Called when the right mouse button was double-clicked.

        From the wxPython demo code:-

        'I do this because I don't like the default behaviour of not starting
        the cell editor on double clicks, but only a second click.'

        Fair enuff!

        """
        row, col = evt.GetRow(), evt.GetCol()

        # try to find a renderer for this cell
        renderer = self.model.get_cell_renderer(row, col)

        # if the renderer has defined a handler then call it
        if renderer is not None:
            renderer.on_right_dclick(self, row, col)

    def _on_cell_right_click(self, evt):
        """ Called when a right click occurred in a cell. """
        row, col = evt.GetRow(), evt.GetCol()

        # try to find a renderer for this cell
        renderer = self.model.get_cell_renderer(row, col)

        # if the renderer has defined a handler then call it
        result = False
        if renderer is not None:
            result = renderer.on_right_click(self, row, col)

        # if the handler didn't tell us to stop further handling then skip
        if not result:
            # ask model for the appropriate context menu
            menu_manager = self.model.get_cell_context_menu(row, col)
            # get the underlying menu object
            if menu_manager is not None:
                controller = None
                if type(menu_manager) is tuple:
                    menu_manager, controller = menu_manager
                menu = menu_manager.create_menu(self._grid, controller)
                # if it has anything in it pop it up
                if menu.GetMenuItemCount() > 0:
                    # Popup the menu (if an action is selected it will be
                    # performed before before 'PopupMenu' returns).
                    x, y = evt.GetPosition().Get()
                    self._grid.PopupMenu(menu, x - 10, y - 10)

            self.cell_right_clicked = (row, col)

            evt.Skip()

    def _on_label_right_click(self, evt):
        """ Called when a right click occurred on a label. """

        row, col = evt.GetRow(), evt.GetCol()

        # a row value of -1 means this click happened on a column.
        # vice versa, a col value of -1 means a row click.
        menu_manager = None
        if row == -1:
            menu_manager = self.model.get_column_context_menu(col)
        else:
            menu_manager = self.model.get_row_context_menu(row)

        if menu_manager is not None:
            # get the underlying menu object
            menu = menu_manager.create_menu(self._grid)
            # if it has anything in it pop it up
            if menu.GetMenuItemCount() > 0:
                # Popup the menu (if an action is selected it will be performed
                # before before 'PopupMenu' returns).
                self._grid.PopupMenu(menu, evt.GetPosition().Get())
        elif col >= 0:
            cws = getattr(self, "_cached_widths", None)
            if (cws is not None) and (0 <= col < len(cws)):
                cws[col] = None
                self.__autosize()

        evt.Skip()

    def _on_label_left_click(self, evt):
        """ Called when a left click occurred on a label. """

        row, col = evt.GetRow(), evt.GetCol()

        # A row value of -1 means this click happened on a column.
        # vice versa, a col value of -1 means a row click.
        if (row == -1) and self.allow_column_sort and evt.ControlDown():
            self._column_sort(col)

        elif (col == -1) and self.allow_row_sort and evt.ControlDown():
            self._row_sort(row)

        evt.Skip()

    def _column_sort(self, col):
        """ Sorts the data on the specified column **col**.
        """
        self._grid.Freeze()

        if (col == self._current_sorted_col) and (not self._col_sort_reversed):
            # If this column is currently sorted on then reverse it:
            self.model.sort_by_column(col, True)
        elif (col == self._current_sorted_col) and self._col_sort_reversed:
            # If this column is currently reverse-sorted then unsort it:
            try:
                self.model.no_column_sort()
            except NotImplementedError:
                # If an unsort function is not implemented then just
                # reverse the sort:
                self.model.sort_by_column(col, False)
        else:
            # Sort the data:
            self.model.sort_by_column(col, False)

        self._grid.Thaw()

    def _row_sort(self, row):
        self._grid.Freeze()

        if (row == self._current_sorted_row) and (not self._row_sort_reversed):
            # If this row is currently sorted on then reverse it:
            self.model.sort_by_row(row, True)
        elif row == self._current_sorted_row and self._row_sort_reversed:
            # If this row is currently reverse-sorted then unsort it:
            try:
                self.model.no_row_sort()
            except NotImplementedError:
                # If an unsort function is not implemented then just
                # reverse the sort:
                self.model.sort_by_row(row, False)
        else:
            # Sort the data:
            self.model.sort_by_row(row, False)

        self._grid.Thaw()

    def _on_key_down(self, evt):
        """ Called when a key is pressed. """
        # This changes the behaviour of the <Enter> and <Tab> keys to make
        # manual data entry smoother!
        #
        # Don't change the behavior if the <Control> key is pressed as this
        # has meaning to the edit control.
        key_code = evt.GetKeyCode()

        evt.Skip()

    def _move_to_next_cell(self, expandSelection=False):
        """ Move to the 'next' cell. """

        # Complete the edit on the current cell.
        self._grid.DisableCellEditControl()

        # Try to move to the next column.
        success = self._grid.MoveCursorRight(expandSelection)

        # If the move failed then we must be at the end of a row.
        if not success:
            # Move to the first column in the next row.
            newRow = self._grid.GetGridCursorRow() + 1
            if newRow < self._grid.GetNumberRows():
                self._grid.SetGridCursor(newRow, 0)
                self._grid.MakeCellVisible(newRow, 0)

            else:
                # This would be a good place to add a new row if your app
                # needs to do that.
                pass

        return success

    def _move_to_previous_cell(self, expandSelection=False):
        """ Move to the 'previous' cell. """

        # Complete the edit on the current cell.
        self._grid.DisableCellEditControl()

        # Try to move to the previous column (without expanding the current
        # selection).
        success = self._grid.MoveCursorLeft(expandSelection)

        # If the move failed then we must be at the start of a row.
        if not success:
            # Move to the last column in the previous row.
            newRow = self._grid.GetGridCursorRow() - 1
            if newRow >= 0:
                self._grid.SetGridCursor(
                    newRow, self._grid.GetNumberCols() - 1
                )
                self._grid.MakeCellVisible(
                    newRow, self._grid.GetNumberCols() - 1
                )

    def _refresh(self):
        self._grid.GetParent().Layout()

    def _on_col_move(self, evt):
        """ Called when a column move is taking place.
        """
        self._ignore = True

        # Get the column being moved:
        frm = evt.GetMoveColumn()

        # Get the column to insert it before:
        to = evt.GetBeforeColumn()

        # Tell the model to update its columns:
        if self.model._move_column(frm, to):

            # Modify the grid:
            grid = self._grid
            cols = grid.GetNumberCols()
            widths = [grid.GetColSize(i) for i in range(cols)]
            width = widths[frm]
            del widths[frm]
            to -= frm < to
            widths.insert(to, width)

            grid.BeginBatch()

            grid.ProcessTableMessage(
                GridTableMessage(
                    self._grid_table_base,
                    GRIDTABLE_NOTIFY_COLS_DELETED,
                    frm,
                    1,
                )
            )

            grid.ProcessTableMessage(
                GridTableMessage(
                    self._grid_table_base,
                    GRIDTABLE_NOTIFY_COLS_INSERTED,
                    to,
                    1,
                )
            )

            [grid.SetColSize(i, widths[i]) for i in range(min(frm, to), cols)]

            grid.EndBatch()

    def _on_row_move(self, evt):
        """ Called when a row move is taking place.
        """
        self._ignore = True

        # Get the column being moved:
        frm = evt.GetMoveRow()

        # Get the column to insert it before:
        to = evt.GetBeforeRow()

        # Tell the model to update its rows:
        if self.model._move_row(frm, to):

            # Notify the grid:
            grid = self._grid
            rows = grid.GetNumberRows()
            heights = [grid.GetRowSize(i) for i in range(rows)]
            height = heights[frm]
            del heights[frm]
            to -= frm < to
            heights.insert(to, height)

            grid.BeginBatch()

            grid.ProcessTableMessage(
                GridTableMessage(
                    self._grid_table_base,
                    GRIDTABLE_NOTIFY_ROWS_DELETED,
                    frm,
                    1,
                )
            )

            grid.ProcessTableMessage(
                GridTableMessage(
                    self._grid_table_base,
                    GRIDTABLE_NOTIFY_ROWS_INSERTED,
                    to,
                    1,
                )
            )

            [grid.SetRowSize(i, heights[i]) for i in range(min(frm, to), rows)]

            grid.EndBatch()

    # ------------------------------------------------------------------------
    # PythonDropTarget interface.
    # ------------------------------------------------------------------------
    def wx_dropped_on(self, x, y, drag_object, drag_result):

        # first resolve the x/y coords into a grid row/col
        row, col = self.__resolve_grid_coords(x, y)

        result = wx.DragNone
        if row != -1 and col != -1:
            # now ask the model if the target cell can accept this object
            valid_target = self.model.is_valid_cell_value(
                row, col, drag_object
            )
            # if this is a valid target then attempt to set the value
            if valid_target:
                # find the data
                data = drag_object
                # sometimes a 'node' attribute on the clipboard gets set
                # to a binding. if this happens we want to use it, otherwise
                # we want to just use the drag_object passed to us
                if (
                    hasattr(enClipboard, "node")
                    and enClipboard.node is not None
                ):
                    data = enClipboard.node

                # now make sure the value gets set in the model
                self.model.set_value(row, col, data)
                result = drag_result

        return result

    def wx_drag_over(self, x, y, drag_object, drag_result):

        # first resolve the x/y coords into a grid row/col
        row, col = self.__resolve_grid_coords(x, y)

        result = wx.DragNone
        if row != -1 and col != -1:
            # now ask the model if the target cell can accept this object
            valid_target = self.model.is_valid_cell_value(
                row, col, drag_object
            )
            if valid_target:
                result = drag_result

        return result

    # ------------------------------------------------------------------------
    # private interface.
    # ------------------------------------------------------------------------

    def __initialize_fonts(self):
        """ Initialize the label fonts. """

        self._on_default_label_font_changed()
        self._on_default_cell_font_changed()
        self._on_default_cell_text_color_changed()
        self._on_grid_line_color_changed()

        self._grid.SetColLabelAlignment(wx.ALIGN_CENTRE, wx.ALIGN_CENTRE)
        self._grid.SetRowLabelAlignment(wx.ALIGN_RIGHT, wx.ALIGN_CENTRE)

    def __initialize_rows(self, model):
        """ Initialize the row headers. """

        # should we really be doing this?
        for row in range(model.get_row_count()):
            if model.is_row_read_only(row):
                attr = wx.grid.GridCellAttr()
                attr.SetReadOnly()
                # attr.SetRenderer(None)
                # attr.SetBackgroundColour('linen')
                self._grid.SetRowAttr(row, attr)

    def __initialize_columns(self, model):
        """ Initialize the column headers. """

        # should we really be doing this?
        for column in range(model.get_column_count()):
            if model.is_column_read_only(column):
                attr = wx.grid.GridCellAttr()
                attr.SetReadOnly()
                # attr.SetRenderer(None)
                # attr.SetBackgroundColour('linen')
                self._grid.SetColAttr(column, attr)

    def __initialize_counts(self, model):
        """ Initializes the row and column counts. """

        if model is not None:
            self._row_count = model.get_row_count()
        else:
            self._row_count = 0

        if model is not None:
            self._col_count = model.get_column_count()
        else:
            self._col_count = 0

    def __initialize_sort_state(self):
        """ Initializes the row and column counts. """

        self._current_sorted_col = None
        self._current_sorted_row = None
        self._col_sort_reversed = False
        self._row_sort_reversed = False

    def __initialize_style_settings(self):

        # make sure all the handlers for traits defining styles get called
        self._on_enable_lines_changed()
        self._on_read_only_changed()
        self._on_selection_mode_changed()
        self._on_column_label_height_changed()
        self._on_row_label_width_changed()
        self._on_show_column_headers_changed()
        self._on_show_row_headers_changed()
        self._on_default_cell_bg_color_changed()
        self._on_default_label_bg_color_changed()
        self._on_default_label_text_color_changed()
        self._on_selection_bg_color_changed()
        self._on_selection_text_color_changed()

    def __get_drag_value(self):
        """ Calculates the drag value based on the current selection. """
        return self.model.get_cell_drag_value(
            self._grid.GetGridCursorRow(), self._grid.GetGridCursorCol()
        )

    def __get_selection(self):
        """ Returns a list of values for the current selection. """

        rows, cols = self.__get_selected_rows_and_cols()

        if len(rows) > 0:
            rows.sort()
            value = self.model.get_rows_selection_value(rows)
        elif len(cols) > 0:
            cols.sort()
            value = self.model.get_cols_selection_value(cols)
        else:
            # our final option -- grab the cell that the cursor is currently in
            row = self._grid.GetGridCursorRow()
            col = self._grid.GetGridCursorCol()
            value = self.model.get_cell_selection_value(row, col)
            if value is not None:
                value = [value]

        if value is None:
            value = []

        return value

    def __get_selected_rows_and_cols(self):
        """ Return lists of the selected rows and the selected columns. """

        # note: because the wx grid control is so screwy, we have limited
        # the selection behavior. we only allow single cells to be selected,
        # or whole row, or whole columns.

        rows = self._grid.GetSelectedRows()
        cols = self._grid.GetSelectedCols()

        # because of wx we have to check this as well
        # note that all this malarkey is working on the assumption that
        # only entire rows or entire columns or single cells are selected.
        top_left = self._grid.GetSelectionBlockTopLeft()
        bottom_right = self._grid.GetSelectionBlockBottomRight()
        selection_mode = self._grid.GetSelectionMode()

        if selection_mode == wxGrid.SelectRows:
            # handle rows differently. figure out which rows were
            # selected. turns out that in this case, wx adds a "block"
            # per row, so we have to cycle over the list returned by
            # the GetSelectionBlock routines
            for i in range(len(top_left)):
                top_point = top_left[i]
                bottom_point = bottom_right[i]
                for row_index in range(top_point[0], bottom_point[0] + 1):
                    rows.append(row_index)
        elif selection_mode == wxGrid.SelectColumns:
            # again, in this case we know that only whole columns can be
            # selected
            for i in range(len(top_left)):
                top_point = top_left[i]
                bottom_point = bottom_right[i]
                for col_index in range(top_point[1], bottom_point[1] + 1):
                    cols.append(col_index)
        elif selection_mode == wxGrid.SelectCells:
            # this is the case where the selection_mode is cell, which also
            # allows complete columns or complete rows to be selected.

            # first find the size of the grid
            row_size = self.model.get_row_count()
            col_size = self.model.get_column_count()

            for i in range(len(top_left)):
                top_point = top_left[i]
                bottom_point = bottom_right[i]
                # precalculate whether this is a row or column select
                row_select = (
                    top_point[1] == 0 and bottom_point[1] == col_size - 1
                )
                col_select = (
                    top_point[0] == 0 and bottom_point[0] == row_size - 1
                )

                if row_select:
                    for row_index in range(top_point[0], bottom_point[0] + 1):
                        rows.append(row_index)
                if col_select:
                    for col_index in range(top_point[1], bottom_point[1] + 1):
                        cols.append(top_point[0])

        return (rows, cols)

    def __fire_selection_changed(self):
        self.selection_changed = True

    def __autosize(self):
        """ Autosize the grid with appropriate flags. """

        model = self.model
        grid = self._grid
        if grid is not None and self.autosize:
            grid.AutoSizeColumns(False)
            grid.AutoSizeRows(False)

        # Whenever we size the grid we need to take in to account any
        # explicitly set column sizes:

        grid.BeginBatch()

        dx, dy = grid.GetClientSize().Get()
        n = model.get_column_count()
        pdx = 0
        wdx = 0.0
        widths = []
        cached = getattr(self, "_cached_widths", None)
        current = [grid.GetColSize(i) for i in range(n)]
        if (cached is None) or (len(cached) != n):
            self._cached_widths = cached = [None] * n

        for i in range(n):
            cw = cached[i]
            if (
                (cw is None)
                or (-cw == current[i])
                or
                # hack: For some reason wx always seems to adjust column 0 by
                # 1 pixel from what we set it to (at least on Windows), so we
                # need to add a little fudge factor just for this column:
                ((i == 0) and (abs(current[i] + cw) <= 1))
            ):
                width = model.get_column_size(i)
                if width is None:
                    width = 0.0
                if width <= 0.0:
                    width = 0.1
                if width <= 1.0:
                    wdx += width
                    cached[i] = -1
                else:
                    width = int(width)
                    pdx += width
                    if cw is None:
                        cached[i] = width
            else:
                cached[i] = width = current[i]
                pdx += width

            widths.append(width)

        # The '-1' below adjusts for an off by 1 error in the way the wx.Grid
        # control determines whether or not it needs a horizontal scroll bar:
        adx = max(0, dx - pdx - 1)

        for i in range(n):
            width = cached[i]
            if width < 0:
                width = widths[i]
                if width <= 1.0:
                    w = max(30, int(round((adx * width) / wdx)))
                    wdx -= width
                    width = w
                    adx -= width
                    cached[i] = -w

            grid.SetColSize(i, width)

        grid.AdjustScrollbars()
        grid.EndBatch()
        grid.ForceRefresh()

    def __resolve_grid_coords(self, x, y):
        """ Resolve the specified x and y coordinates into row/col
            coordinates. Returns row, col. """

        # the x,y coordinates here are Unscrolled coordinates.
        # They must be changed to scrolled coordinates.
        x, y = self._grid.CalcUnscrolledPosition(x, y)

        # now we need to get the row and column from the grid
        # but we need to first remove the RowLabel and ColumnLabel
        # bounding boxes
        if self.show_row_headers:
            x = x - self._grid.GetGridRowLabelWindow().GetRect().width

        if self.show_column_headers:
            y = y - self._grid.GetGridColLabelWindow().GetRect().height

        return (self._grid.YToRow(y), self._grid.XToCol(x))

    def _select_rows(self, cells):
        """ Selects all of the rows specified by a list of (row,column) pairs.
        """
        # For a large set of rows, simply calling 'SelectBlock' on the Grid
        # object for each row is very inefficient, so we first make a pass over
        # all of the cells to merge them into contiguous ranges as much as
        # possible:
        sb = self._grid.SelectBlock

        # Extract the rows and sort them:
        rows = [row for row, column in cells]
        rows.sort()

        # Now find contiguous ranges of rows, and select the current range
        # whenever a break in the sequence is found:
        first = last = -999
        for row in rows:
            if row == (last + 1):
                last = row
            else:
                if first >= 0:
                    sb(first, 0, last, 0, True)
                first = last = row

        # Handle the last pending range of lines to be selected:
        if first >= 0:
            sb(first, 0, last, 0, True)


class _GridTableBase(GridTableBase):
    """ A private adapter for the underlying wx grid implementation. """

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __init__(self, model, grid):
        """ Creates a new table base. """

        # Base class constructor.
        GridTableBase.__init__(self)

        # The Pyface model that provides the data.
        self.model = model
        self._grid = grid

        # hacky state variables so we can identify when rows have been
        # added or deleted.
        self._row_count = -1
        self._col_count = -1

        # caches for editors and renderers
        self._editor_cache = {}
        self._renderer_cache = {}

    def dispose(self):

        # Make sure dispose gets called on all traits editors:
        for editor in self._editor_cache.values():
            editor.dispose()
        self._editor_cache = {}

        for renderer in self._renderer_cache.values():
            renderer.dispose()
        self._renderer_cache = {}

    # ------------------------------------------------------------------------
    # 'wxGridTableBase' interface.
    # ------------------------------------------------------------------------

    def GetNumberRows(self):
        """ Return the number of rows in the model. """

        # because wx is such a wack job we have to store away the row
        # and column counts so we can figure out when rows or cols have
        # been appended or deleted. lacking a better place to do this
        # we just set the local variable every time GetNumberRows is called.
        self._row_count = self.model.get_row_count()
        return self._row_count

    def GetNumberCols(self):
        """ Return the number of columns in the model. """

        # see comment in GetNumberRows for logic here
        self._col_count = self.model.get_column_count()
        return self._col_count

    def IsEmptyCell(self, row, col):
        """ Is the specified cell empty? """

        return self.model.is_cell_empty(row, col)

    def GetValue(self, row, col):
        """ Get the value at the specified row and column. """

        return self.model.get_value(row, col)

    def SetValue(self, row, col, value):
        """ Set the value at the specified row and column. """
        return self.model.set_value(row, col, value)

    def GetRowLabelValue(self, row):
        """ Called when the grid needs to display a row label. """

        label = self.model.get_row_name(row)

        if row == self._grid._current_sorted_row:
            if self._grid._row_sort_reversed:
                if is_win32:
                    ulabel = str(label, "ascii") + "  \u00ab"
                    label = ulabel.encode("latin-1")
                else:
                    label += "  <<"
            elif is_win32:
                ulabel = str(label, "ascii") + "  \u00bb"
                label = ulabel.encode("latin-1")
            else:
                label += "  >>"

        return label

    def GetColLabelValue(self, col):
        """ Called when the grid needs to display a column label. """

        label = self.model.get_column_name(col)

        if col == self._grid._current_sorted_col:
            if self._grid._col_sort_reversed:
                if is_win32:
                    ulabel = str(label, "ascii") + "  \u00ab"
                    label = ulabel.encode("latin-1")
                else:
                    label += "  <<"
            elif is_win32:
                ulabel = str(label, "ascii") + "  \u00bb"
                label = ulabel.encode("latin-1")
            else:
                label += "  >>"

        return label

    def GetTypeName(self, row, col):
        """ Called to determine the kind of editor/renderer to use.

        This doesn't necessarily have to be the same type used natively by the
        editor/renderer if they know how to convert.

        """
        typename = None
        try:
            typename = self.model.get_type(row, col)
        except NotImplementedError:
            pass

        if typename is None:
            typename = GRID_VALUE_STRING

        return typename

    def DeleteRows(self, pos, num_rows):
        """ Called when the view is deleting rows. """

        # clear the cache
        self._clear_cache()
        return self.model.delete_rows(pos, num_rows)

    def InsertRows(self, pos, num_rows):
        """ Called when the view is inserting rows. """
        # clear the cache
        self._clear_cache()
        return self.model.insert_rows(pos, num_rows)

    def AppendRows(self, num_rows):
        """ Called when the view is inserting rows. """
        # clear the cache
        self._clear_cache()
        pos = self.model.get_row_count()
        return self.model.insert_rows(pos, num_rows)

    def DeleteCols(self, pos, num_cols):
        """ Called when the view is deleting columns. """

        # clear the cache
        self._clear_cache()
        return self.model.delete_columns(pos, num_cols)

    def InsertCols(self, pos, num_cols):
        """ Called when the view is inserting columns. """
        # clear the cache
        self._clear_cache()
        return self.model.insert_columns(pos, num_cols)

    def AppendCols(self, num_cols):
        """ Called when the view is inserting columns. """
        # clear the cache
        self._clear_cache()
        pos = self.model.get_column_count()
        return self.model.insert_columns(pos, num_cols)

    def GetAttr(self, row, col, kind):
        """ Retrieve the cell attribute object for the specified cell. """

        result = GridCellAttr()
        # we only handle cell requests, for other delegate to the supa
        if kind != GridCellAttr.Cell and kind != GridCellAttr.Any:
            return result

        rows = self.model.get_row_count()
        cols = self.model.get_column_count()

        # First look in the cache for the editor:
        editor = self._editor_cache.get((row, col))
        if editor is None:
            if (row >= rows) or (col >= cols):
                editor = DummyGridCellEditor()
            else:
                # Ask the underlying model for an editor for this cell:
                editor = self.model.get_cell_editor(row, col)
                if editor is not None:
                    self._editor_cache[(row, col)] = editor
                    editor._grid_info = (self._grid._grid, row, col)

        if False:  # editor is not None:
            # Note: We have to increment the reference to keep the
            #       underlying code from destroying our object.
            editor.IncRef()
            result.SetEditor(editor)

        # try to find a renderer for this cell
        renderer = None
        if row < rows and col < cols:
            renderer = self.model.get_cell_renderer(row, col)

        if renderer is not None and renderer.renderer is not None:
            renderer.renderer.IncRef()
            result.SetRenderer(renderer.renderer)

        # look to see if this cell is editable
        read_only = False
        if row < rows and col < cols:
            read_only = (
                self.model.is_cell_read_only(row, col)
                or self.model.is_row_read_only(row)
                or self.model.is_column_read_only(col)
            )

        result.SetReadOnly(read_only)
        if read_only:
            read_only_color = self._grid.default_cell_read_only_color
            if (
                read_only_color is not None
                and read_only_color is not Undefined
            ):
                result.SetBackgroundColour(read_only_color)

        # check to see if colors or fonts are specified for this cell
        bgcolor = None
        if row < rows and col < cols:
            bgcolor = self.model.get_cell_bg_color(row, col)
        else:
            bgcolor = self._grid.default_cell_bg_color

        if bgcolor is not None:
            result.SetBackgroundColour(bgcolor)

        text_color = None
        if row < rows and col < cols:
            text_color = self.model.get_cell_text_color(row, col)
        else:
            text_color = self._grid.default_cell_text_color
        if text_color is not None:
            result.SetTextColour(text_color)

        cell_font = None
        if row < rows and col < cols:
            cell_font = self.model.get_cell_font(row, col)
        else:
            cell_font = self._grid.default_cell_font
        if cell_font is not None:
            result.SetFont(cell_font)

        # check for alignment definition for this cell
        halignment = valignment = None
        if row < rows and col < cols:
            halignment = self.model.get_cell_halignment(row, col)
            valignment = self.model.get_cell_valignment(row, col)
        if halignment is not None and valignment is not None:
            if halignment == "center":
                h = wx.ALIGN_CENTRE
            elif halignment == "right":
                h = wx.ALIGN_RIGHT
            else:
                h = wx.ALIGN_LEFT

            if valignment == "top":
                v = wx.ALIGN_TOP
            elif valignment == "bottom":
                v = wx.ALIGN_BOTTOM
            else:
                v = wx.ALIGN_CENTRE

            result.SetAlignment(h, v)

        return result

    # ------------------------------------------------------------------------
    # private interface.
    # ------------------------------------------------------------------------
    def _clear_cache(self):
        """ Clean out the editor/renderer cache. """

        # Dispose of the editors in the cache after a brief delay, so as
        # to allow completion of the current event:
        do_later(self._editor_dispose, list(self._editor_cache.values()))

        self._editor_cache = {}
        self._renderer_cache = {}

    def _editor_dispose(self, editors):
        for editor in editors:
            editor.dispose()


from wx.grid import GridCellEditor


class DummyGridCellEditor(GridCellEditor):
    def Show(self, show, attr):
        return



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
"""
This module defines a :py:class:`GUIApplication` subclass of
:py:class:`pyface.application.Application`.  This adds cross-platform GUI
application support to the base class via :py:class:`pyface.application.GUI`.

At a minimum this class expects to be provided with a factory that returns
:py:class:`pyface.i_window.IWindow` instances.  For pure Pyface applications
this is most likely to be a subclass of
:py:class:`pyface.application_window.ApplicationWindow`.

"""


import logging

from traits.api import (
    Bool,
    Callable,
    Instance,
    List,
    ReadOnly,
    Tuple,
    Undefined,
    Vetoable,
    observe,
)

from .application import Application
from .i_dialog import IDialog
from .i_splash_screen import ISplashScreen
from .i_window import IWindow
from .ui_traits import Image

logger = logging.getLogger(__name__)


def default_window_factory(application, **kwargs):
    """ The default window factory returns an application window.

    This is almost never the right thing, but allows users to get off the
    ground with the base class.
    """
    from pyface.application_window import ApplicationWindow

    return ApplicationWindow(**kwargs)


class GUIApplication(Application):
    """ A basic Pyface GUI application. """

    # 'GUIApplication' traits -------------------------------------------------

    # Branding ---------------------------------------------------------------

    #: The splash screen for the application. No splash screen by default
    splash_screen = Instance(ISplashScreen)

    #: The about dialog for the application.
    about_dialog = Instance(IDialog)

    #: Icon for the application (used in window titlebars)
    icon = Image

    #: Logo of the application (used in splash screens and about dialogs)
    logo = Image

    # Window management ------------------------------------------------------

    #: The window factory to use when creating a window for the application.
    window_factory = Callable(default_window_factory)

    #: Default window size
    window_size = Tuple((800, 600))

    #: Currently active Window if any
    active_window = Instance(IWindow)

    #: List of all open windows in the application
    windows = List(Instance(IWindow))

    #: The Pyface GUI instance for the application
    gui = ReadOnly

    # Protected interface ----------------------------------------------------

    #: Flag if the exiting of the application was explicitely requested by user
    # An 'explicit' exit is when the 'exit' method is called.
    # An 'implicit' exit is when the user closes the last open window.
    _explicit_exit = Bool(False)

    # -------------------------------------------------------------------------
    # 'GUIApplication' interface
    # -------------------------------------------------------------------------

    # Window lifecycle methods -----------------------------------------------

    def create_window(self, **kwargs):
        """ Create a new application window.

        By default uses the :py:attr:`window_factory` to do this.  Subclasses
        can override if they want to do something different or additional.

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments to pass to the window factory.

        Returns
        -------
        window : IWindow instance or None
            The new IWindow instance.
        """
        window = self.window_factory(application=self, **kwargs)

        if window.size == (-1, -1):
            window.size = self.window_size
        if not window.title:
            window.title = self.name
        if self.icon:
            window.icon = self.icon

        return window

    def add_window(self, window):
        """ Add a new window to the windows we are tracking. """

        # Keep a handle on all windows created so that non-active windows don't
        # get garbage collected
        self.windows.append(window)

        # Something might try to veto the opening of the window.
        opened = window.open()
        if opened:
            window.activate()

    # Action handlers --------------------------------------------------------

    def do_about(self):
        """ Display the about dialog, if it exists. """
        if self.about_dialog is not None:
            self.about_dialog.open()

    # -------------------------------------------------------------------------
    # 'Application' interface
    # -------------------------------------------------------------------------

    def start(self):
        """ Start the application, setting up things that are required

        Subclasses should open at least one ApplicationWindow or subclass in
        their start method, and should call the superclass start() method
        before doing any work themselves.
        """
        from pyface.gui import GUI

        ok = super(GUIApplication, self).start()
        if ok:
            # create the GUI so that the splash screen comes up first thing
            if self.gui is Undefined:
                self.gui = GUI(splash_screen=self.splash_screen)

            # create the initial windows to show
            self._create_windows()

        return ok

    # -------------------------------------------------------------------------
    # 'GUIApplication' Private interface
    # -------------------------------------------------------------------------

    def _create_windows(self):
        """ Create the initial windows to display.

        By default calls :py:meth:`create_window` once. Subclasses can
        override this method.
        """
        window = self.create_window()
        self.add_window(window)

    # -------------------------------------------------------------------------
    # 'Application' private interface
    # -------------------------------------------------------------------------

    def _run(self):
        """ Actual implementation of running the application: starting the GUI
        event loop.
        """
        # Fire a notification that the app is running.  This is guaranteed to
        # happen after all initialization has occurred and the event loop has
        # started.  A listener for this event is a good place to do things
        # where you want the event loop running.
        self.gui.invoke_later(
            self._fire_application_event, "application_initialized"
        )

        # start the GUI - script blocks here
        self.gui.start_event_loop()
        return True

    # Destruction methods -----------------------------------------------------

    def _can_exit(self):
        """ Check with each window to see if it can be closed

        The fires closing events for each window, and returns False if any
        listener vetos.
        """
        if not super(GUIApplication, self)._can_exit():
            return False

        for window in reversed(self.windows):
            window.closing = event = Vetoable()
            if event.veto:
                return False
        else:
            return True

    def _prepare_exit(self):
        """ Close each window """
        # ensure copy of list, as we modify original list while closing
        for window in list(reversed(self.windows)):
            window.destroy()
            window.closed = window

    def _exit(self):
        """ Shut down the event loop """
        self.gui.stop_event_loop()

    # Trait default handlers ------------------------------------------------

    def _window_factory_default(self):
        """ Default to ApplicationWindow

        This is almost never the right thing, but allows users to get off the
        ground with the base class.
        """
        from pyface.application_window import ApplicationWindow

        return lambda application, **kwargs: ApplicationWindow(**kwargs)

    def _splash_screen_default(self):
        """ Default SplashScreen """
        from pyface.splash_screen import SplashScreen

        dialog = SplashScreen()
        if self.logo:
            dialog.image = self.logo
        return dialog

    def _about_dialog_default(self):
        """ Default AboutDialog """
        from html import escape
        
        from pyface.about_dialog import AboutDialog

        additions = [
            "<h1>{}</h1>".format(escape(self.name)),
            "Copyright &copy; 2018 {}, all rights reserved".format(
                escape(self.company)
            ),
            "",
        ]
        additions += [escape(line) for line in self.description.split("\n\n")]

        dialog = AboutDialog(
            title="About {}".format(self.name), additions=additions
        )
        if self.logo:
            dialog.image = self.logo
        return dialog

    # Trait listeners --------------------------------------------------------

    @observe("windows:items:activated")
    def _on_activate_window(self, event):
        """ Listener that tracks currently active window.
        """
        window = event.object
        if window in self.windows:
            self.active_window = window

    @observe("windows:items:deactivated")
    def _on_deactivate_window(self, event):
        """ Listener that tracks currently active window.
        """
        self.active_window = None

    @observe("windows:items:closed")
    def _on_window_closed(self, event):
        """ Listener that ensures window handles are released when closed.
        """
        window = event.object
        if window in self.windows:
            self.windows.remove(window)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from codecs import decode, encode
from functools import partial
from typing import Any as TAny, Callable as TCallable, NamedTuple

from traits.api import Any, HasStrictTraits, Interface


class DataFormat(NamedTuple):
    """ Information about a mimetype and serializers.

    Simple namedtuple-based class that stores the mimetype, serializer and
    deserializer together.
    """

    #: The mimetype of the data.
    mimetype: str

    #: A callable that serializes this format.  It should take an python
    #: object of a supported type, and return a bytestring.
    serialize: TCallable[[TAny], bytes]

    #: A callable that deserializes this format.  It should take a
    #: bytestring and return the extracted object.
    deserialize: TCallable[[bytes], TAny]


def text_format(encoding='utf-8', mimetype='text/plain'):
    """ DataFormat factory for text mimetypes.
    """
    return DataFormat(
        mimetype=mimetype,
        serialize=partial(encode, encoding=encoding),
        deserialize=partial(decode, encoding=encoding),
    )


class IDataWrapper(Interface):
    """ Wrapper around polymorphic toolkit data object containing mimedata.

    To support clipboard and drag and drop operations, toolkits need a way
    of generically representing data in multiple formats.  This is a wrapper
    class that provides a toolkit independent intreface to these classes
    which allow the exchange of data, with types specified by MIME types.
    """

    #: The toolkit data object.
    toolkit_data = Any()

    def mimetypes(self):
        """ Return a set of mimetypes holding data.

        Returns
        -------
        mimetypes : set of str
            The set of mimetypes currently storing data in the toolkit data
            object.
        """
        pass

    def has_format(self, format):
        """ Whether or not a particular format has available data.

        Parameters
        ----------
        format : DataFormat
            A data format object.

        Returns
        -------
        has_format : bool
            Whether or not there is data associated with that format in the
            underlying toolkit object.
        """
        raise NotImplementedError()

    def get_format(self, format):
        """ The decoded data associted with the format.

        Parameters
        ----------
        format : DataFormat
            A data format object.

        Returns
        -------
        data : any
            The data decoded for the given format.
        """
        raise NotImplementedError()

    def set_format(self, format, data):
        """ Encode and set data for the format.

        Parameters
        ----------
        format : DataFormat
            A data format object.
        data : any
            The data to be encoded and stored.
        """
        raise NotImplementedError()

    def get_mimedata(self, mimetype):
        """ Get raw data for the given media type.

        Parameters
        ----------
        mimetype : str
            The mime media type to be extracted.

        Returns
        -------
        mimedata : bytes
            The mime media data as bytes.
        """
        raise NotImplementedError()

    def set_mimedata(self, mimetype, mimedata):
        """ Set raw data for the given media type.

        Parameters
        ----------
        mimetype : str
            The mime media type to be extracted.
        mimedata : bytes
            The mime media data encoded as bytes..
        """
        raise NotImplementedError()


class MDataWrapper(HasStrictTraits):
    """ Mixin class for DataWrappers.

    This provides standard methods for using DataFormat objects, but not the
    low-level communication with the underlying toolkit.
    """

    def has_format(self, format):
        """ Whether or not a particular format has available data.

        Parameters
        ----------
        format : DataFormat
            A data format object.

        Returns
        -------
        has_format : bool
            Whether or not there is data associated with that format in the
            underlying toolkit object.
        """
        return format.mimetype in self.mimetypes()

    def get_format(self, format):
        """ The decoded data associted with the format.

        Parameters
        ----------
        format : DataFormat
            A data format object.

        Returns
        -------
        data : any
            The data decoded for the given format.
        """
        return format.deserialize(self.get_mimedata(format.mimetype))

    def set_format(self, format, data):
        """ Encode and set data for the format.

        Parameters
        ----------
        format : DataFormat
            A data format object.
        data : any
            The data to be encoded and stored.
        """
        self.set_mimedata(format.mimetype, format.serialize(data))



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Default base-class for resource factories. """


class ResourceFactory(object):
    """ Default base-class for resource factories. """

    # ------------------------------------------------------------------------
    # 'ResourceFactory' interface.
    # ------------------------------------------------------------------------

    def image_from_file(self, filename):
        """ Creates an image from the data in the specified filename. """

        raise NotImplementedError()

    def image_from_data(self, data):
        """ Creates an image from the specified data. """

        raise NotImplementedError()



# (C) Copyright 2008-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# ------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought undo package component>
# ------------------------------------------------------------------------------

# Enthought library imports.
from traits.api import (
    Bool,
    Event,
    HasTraits,
    Instance,
    Int,
    Property,
    Str,
    provides,
)

# Local imports.
from .i_undo_manager import IUndoManager


@provides(IUndoManager)
class UndoManager(HasTraits):
    """The UndoManager class is the default implementation of the
    IUndoManager interface.
    """

    #### 'IUndoManager' interface #############################################

    #: This is the currently active command stack and may be None. Typically it
    #: is set when some sort of editor becomes active.
    active_stack = Instance("pyface.undo.api.ICommandStack")

    #: This reflects the clean state of the currently active command stack. It
    #: is intended to support a "document modified" indicator in the GUI. It is
    #: maintained by the undo manager.
    active_stack_clean = Property(Bool)

    #: This is the name of the command that can be redone. It will be empty if
    #: there is no command that can be redone. It is maintained by the undo
    #: manager.
    redo_name = Property(Str)

    #: This is the sequence number of the next command to be performed. It is
    #: incremented immediately before a command is invoked (by its 'do()'
    #: method).
    sequence_nr = Int()

    #: This event is fired when the index of a command stack changes. The value
    #: of the event is the stack that has changed. Note that it may not be the
    #: active stack.
    stack_updated = Event()

    #: This is the name of the command that can be undone. It will be empty if
    #: there is no command that can be undone. It is maintained by the undo
    #: manager.
    undo_name = Property(Str)

    ###########################################################################
    # 'IUndoManager' interface.
    ###########################################################################

    def redo(self):
        """ Redo the last undone command of the active command stack. """

        if self.active_stack is not None:
            self.active_stack.redo()

    def undo(self):
        """ Undo the last command of the active command stack. """

        if self.active_stack is not None:
            self.active_stack.undo()

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _active_stack_changed(self, new):
        """ Handle a different stack becoming active. """

        # Pretend that the stack contents have changed.
        self.stack_updated = new

    def _get_active_stack_clean(self):
        """ Get the current clean state. """

        if self.active_stack is None:
            active_stack_clean = True
        else:
            active_stack_clean = self.active_stack.clean

        return active_stack_clean

    def _get_redo_name(self):
        """ Get the current redo name. """

        if self.active_stack is None:
            redo_name = ""
        else:
            redo_name = self.active_stack.redo_name

        return redo_name

    def _get_undo_name(self):
        """ Get the current undo name. """

        if self.active_stack is None:
            undo_name = ""
        else:
            undo_name = self.active_stack.undo_name

        return undo_name



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
import contextlib
from functools import wraps
import operator
import re
from unittest import (
    mock,
    TestSuite,
)

from packaging.version import Version

from traits import __version__ as TRAITS_VERSION


def filter_tests(test_suite, exclusion_pattern):
    filtered_test_suite = TestSuite()
    for item in test_suite:
        if isinstance(item, TestSuite):
            filtered = filter_tests(item, exclusion_pattern)
            filtered_test_suite.addTest(filtered)
        else:
            match = re.search(exclusion_pattern, item.id())
            if match is not None:
                skip_msg = "Test excluded via pattern '{}'".format(
                    exclusion_pattern
                )
                setattr(item, 'setUp', lambda: item.skipTest(skip_msg))
            filtered_test_suite.addTest(item)
    return filtered_test_suite


def has_traitsui():
    """ Is traitsui installed? """
    try:
        import traitsui
    except ImportError:
        return False
    return True


def skip_if_no_traitsui(test):
    """ Decorator that skips test if traitsui not available """

    @wraps(test)
    def new_test(self):
        if has_traitsui():
            test(self)
        else:
            self.skipTest("Can't import traitsui.")

    return new_test


def is_traits_version_ge(version):
    """ Return true if the traits version is greater than or equal to the
    required value.

    Parameters
    ----------
    version : str
        Version to be parsed. e.g. "6.0"
    """
    traits_version = Version(TRAITS_VERSION)
    given_version = Version(version)
    return traits_version >= given_version



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Provides an AbstractDataModel ABC for Pyface data models.

This module provides an ABC for all data view data models.  This specifies
the API that the data view widgets expect, and which the underlying
data is adapted to by the concrete implementations.  Data models are intended
to be toolkit-independent, and be able to adapt any approximately tabular or
nested data structure to what the data view system expects.
"""
from abc import abstractmethod

from traits.api import ABCHasStrictTraits, Event, Instance

from .data_view_errors import DataViewSetError
from .index_manager import AbstractIndexManager


class AbstractDataModel(ABCHasStrictTraits):
    """ Abstract base class for Pyface data models.

    The data model API is intended to provide a common API for hierarchical
    and tabular data.  This class is concerned with the structure, type and
    values provided by the data, but not with how the data is presented.

    Row and column indices are represented by sequences (usually lists) of
    integers, specifying the index at each level of the hierarchy.  The root
    row and column are represented by empty lists.

    Subclasses need to implement the ``get_column_count``,
    ``can_have_children`` and ``get_row_count`` methods to return the number
    of columns in a particular row, as well as the hierarchical structure of
    the rows.  Appropriate observers should be set up on the underlaying data
    so that the ``structure_changed`` event is fired when the values returned
    by these methods would change.

    Subclasses also have to implement the ``get_value`` and ``get_value_type``
    methods.  These expect a row and column index, with root values treated
    specially: the root row corresponds to the values which will be displayed
    in the column headers of the view, and the root column corresponds to the
    values which will be displayed in the row headers of the view.
    The ``get_value`` returns an arbitrary Python object corresponding to the
    cell being viewed, and the ``get_value_type`` should return an instance of
    an ``AbstractValueType`` that adapts the raw value to the data channels
    that the data view expects (eg. text, color, icons, editable value, etc.).
    Implementations should ensure that the ``values_changed`` event fires
    whenever the data, or the way the data is presented, is updated.

    If the data is to be editable then the subclass should override the
    ``set_value`` method.  It should attempt to change the underlying data as a
    side-effect or raise DataViewSetError on failure (for example,
    setting an invalid value).  If the underlying data structure cannot be
    listened to internally (such as a numpy array or Pandas data frame),
    ``set_value`` should also fire the ``values_changed`` event with
    appropriate values.

    In the cases where the underlying data structure cannot be observed by
    the usual traits mechanisms, the end-user of the code may be responsible
    for ensuring that the ``structure_changed`` and ``values_changed`` events
    are fired appropriately.
    """

    #: The index manager that helps convert toolkit indices to data view
    #: indices.  This should be an IntIndexManager for non-hierarchical data
    #: or a TupleIndexManager for hierarchical data.
    index_manager = Instance(AbstractIndexManager)

    #: Event fired when the structure of the data changes.
    structure_changed = Event()

    #: Event fired when value changes without changes to structure.  This
    #: should be set to a 4-tuple of (start_row_index, start_column_index,
    #: end_row_index, end_column_index) indicated the subset of data which
    #: changed.  These end values are inclusive, unlike standard Python
    #: slicing notation.
    values_changed = Event()

    # Data structure methods

    @abstractmethod
    def get_column_count(self):
        """ How many columns in the data view model.

        Returns
        -------
        column_count : non-negative int
            The number of columns that the data view provides.  This count
            should not include the row header.
        """
        raise NotImplementedError()

    @abstractmethod
    def can_have_children(self, row):
        """ Whether or not a row can have child rows.

        The root row should always return True.

        Parameters
        ----------
        row : sequence of int
            The indices of the row as a sequence from root to leaf.

        Returns
        -------
        can_have_children : bool
            Whether or not the row can ever have child rows.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_row_count(self, row):
        """ How many child rows the row currently has.

        Parameters
        ----------
        row : sequence of int
            The indices of the row as a sequence from root to leaf.

        Returns
        -------
        row_count : non-negative int
            The number of child rows that the row has.
        """
        raise NotImplementedError()

    # Data value methods

    @abstractmethod
    def get_value(self, row, column):
        """ Return the Python value for the row and column.

        The values for column headers are returned by calling this method with
        row equal to ().  The values for row headers are returned by calling
        this method with column equal to ().

        Parameters
        ----------
        row : sequence of int
            The indices of the row as a sequence from root to leaf.
        column : sequence of int
            The indices of the column as a sequence of length 0 or 1.

        Returns
        -------
        value : any
            The value represented by the given row and column.

        Raises
        -------
        DataViewGetError
            If the value cannot be accessed in an expected way.  If this is
            raised then the error will be ignored and not logged by the
            data view infrastructure.
        """
        raise NotImplementedError()

    def can_set_value(self, row, column):
        """ Whether the value in the indicated row and column can be set.

        The default method assumes the data is read-only and always
        returns False.

        Whether or a column header can be set is returned by calling this
        method with row equal to ().  Whether or a row header can be set
        is returned by calling this method with column equal to ().

        Parameters
        ----------
        row : sequence of int
            The indices of the row as a sequence from root to leaf.
        column : sequence of int
            The indices of the column as a sequence of length 0 or 1.

        Returns
        -------
        can_set_value : bool
            Whether or not the value can be set.
        """
        return False

    def set_value(self, row, column, value):
        """ Set the Python value for the row and column.

        The default method assumes the data is read-only and always
        returns False.

        The values for column headers can be set by calling this method with
        row equal to ().  The values for row headers can be set by calling
        this method with column equal to ().

        Parameters
        ----------
        row : sequence of int
            The indices of the row as a sequence from root to leaf.
        column : sequence of int
            The indices of the column as a sequence of length 0 or 1.
        value : any
            The new value for the given row and column.

        Raises
        -------
        DataViewSetError
            If the value cannot be set.
        """
        raise DataViewSetError()

    @abstractmethod
    def get_value_type(self, row, column):
        """ Return the value type of the given row and column.

        The value type for column headers are returned by calling this method
        with row equal to ().  The value types for row headers are returned
        by calling this method with column equal to ().

        Parameters
        ----------
        row : sequence of int
            The indices of the row as a sequence from root to leaf.
        column : sequence of int
            The indices of the column as a sequence of length 0 or 1.

        Returns
        -------
        value_type : AbstractValueType or None
            The value type of the given row and column, or None if no value
            should be displayed.
        """
        raise NotImplementedError()

    # Convenience methods

    def is_row_valid(self, row):
        """ Return whether or not the given row index refers to a valid row.

        A row index is valid if every value in the tuple is between 0 and the
        number of child rows of the parent.

        Parameters
        ----------
        row : sequence of int
            The row to check as indices of the row from root to leaf.

        Returns
        -------
        valid : bool
            Whether or not the row index is valid.
        """
        for i, index in enumerate(row):
            parent = row[:i]
            if not self.can_have_children(parent):
                return False
            if not 0 <= index < self.get_row_count(parent):
                return False
        return True

    def is_column_valid(self, column):
        """ Return whether or not the given column index refers to a valid column.

        A column index is valid if it is the root, or the value is between 0 and
        the number of columns in the model.

        Parameters
        ----------
        column :  sequence of int
            The column to check.

        Returns
        -------
        valid : bool
            Whether or not the column index is valid.
        """
        if len(column) == 1:
            return 0 <= column[0] < self.get_column_count()

        return len(column) == 0

    def iter_rows(self, start_row=()):
        """ Iterator that yields rows in preorder.

        Parameters
        ----------
        start_row : sequence of int
            The row to start at.  The iterator will yeild the row and all
            descendant rows.

        Yields
        ------
        row_index : sequence of int
            The current row index.
        """
        start_row = tuple(start_row)
        yield start_row
        if self.can_have_children(start_row):
            for row in range(self.get_row_count(start_row)):
                yield from self.iter_rows(start_row + (row,))

    def iter_items(self, start_row=()):
        """ Iterator that yields rows and columns in preorder.

        This yields pairs of row, column for all rows in preorder
        and and all column indices for all rows, including ().
        Columns are iterated in order.

        Parameters
        ----------
        start_row : sequence of int
            The row to start iteration from.

        Yields
        ------
        row_index, column_index
            The current row and column indices.
        """
        for row in self.iter_rows(start_row):
            yield row, ()
            for column in range(self.get_column_count()):
                yield row, (column,)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""A dialog that is loaded from an XRC resource file.
"""


import os.path

# Major packages.
import wx
import wx.xrc


from traits.api import Instance, Str
import traits.util.resource


from .dialog import Dialog


# ----------------------------------------------------------------------------
# class 'XrcDialog'
# ----------------------------------------------------------------------------


class XrcDialog(Dialog):
    """A dialog that is loaded from an XRC resource file.
    """

    # ------------------------------------------------------------------------
    # Traits
    # ------------------------------------------------------------------------

    # 'XrcDialog' interface --------------------------------------------

    # Path to the xrc file relative to the class's module
    xrc_file = Str()

    # The ID of the dialog in the file
    id = Str("dialog")

    # The resource object
    resource = Instance(wx.xrc.XmlResource)

    # ------------------------------------------------------------------------
    # 'Dialog' interface
    # ------------------------------------------------------------------------

    def _create_control(self, parent):
        """
        Creates the dialog and loads it in from the resource file.
        """
        classpath = traits.util.resource.get_path(self)
        path = os.path.join(classpath, self.xrc_file)

        self.resource = wx.xrc.XmlResource(path)
        return self.resource.LoadDialog(parent, self.id)

    def _create_contents(self, dialog):
        """
        Calls add_handlers.  The actual content is created
        in _create_control by loading a resource file.
        """
        # Wire up the standard buttons
        # We change the ID on OK and CANCEL to the standard ids
        # so we get the default behavior
        okbutton = self.XRCCTRL("OK")
        if okbutton is not None:
            # Change the ID and set the handler
            okbutton.SetId(wx.ID_OK)
            self.control.Bind(wx.EVT_BUTTON, okbutton.GetId(), self._on_ok)
        cancelbutton = self.XRCCTRL("CANCEL")
        if cancelbutton is not None:
            # Change the ID and set the handler
            cancelbutton.SetId(wx.ID_CANCEL)
            self.control.Bind(
                wx.EVT_BUTTON, self._on_cancel, cancelbutton.GetId()
            )
        helpbutton = self.XRCCTRL("HELP")
        if helpbutton is not None:
            self.control.Bind(wx.EVT_BUTTON, self._on_help, helpbutton.GetId())

        self._add_handlers()

    # ------------------------------------------------------------------------
    # 'XrcDialog' interface
    # ------------------------------------------------------------------------

    def XRCID(self, name):
        """
        Returns the numeric widget id for the given name.
        """
        return wx.xrc.XRCID(name)

    def XRCCTRL(self, name):
        """
        Returns the control with the given name.
        """
        return self.control.Window.FindWindowById(self.XRCID(name))

    def set_validator(self, name, validator):
        """
        Sets the validator on the named control.
        """
        self.XRCCTRL(name).SetValidator(validator)

    # ------------------------------------------------------------------------
    # 'XrcDialog' protected interface
    # ------------------------------------------------------------------------

    def _add_handlers(self):
        """
        Override to add event handlers.
        """
        return



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from unittest import TestCase
from unittest.mock import Mock

from pyface.data_view.abstract_value_type import CheckState
from pyface.data_view.data_view_errors import DataViewSetError
from pyface.data_view.value_types.bool_value import BoolValue


class TestBoolValue(TestCase):

    def setUp(self):
        self.model = Mock()
        self.model.get_value = Mock(return_value=True)
        self.model.can_set_value = Mock(return_value=True)
        self.model.set_value = Mock()

    def test_defaults(self):
        value = BoolValue()
        self.assertEqual(value.true_text, "")
        self.assertEqual(value.false_text, "")

    def test_has_text_default(self):
        value = BoolValue()
        has_text = value.has_text(self.model, [0], [0])
        self.assertFalse(has_text)

    def test_has_text(self):
        value = BoolValue(true_text="Yes", false_text="No")
        has_text = value.has_text(self.model, [0], [0])
        self.assertTrue(has_text)

    def test_get_text_default(self):
        value = BoolValue()
        text = value.get_text(self.model, [0], [0])
        self.assertEqual(text, "")

        self.model.get_value = Mock(return_value=False)
        text = value.get_text(self.model, [0], [0])
        self.assertEqual(text, "")

    def test_get_text(self):
        value = BoolValue(true_text="Yes", false_text="No")
        text = value.get_text(self.model, [0], [0])
        self.assertEqual(text, "Yes")

        self.model.get_value = Mock(return_value=False)
        text = value.get_text(self.model, [0], [0])
        self.assertEqual(text, "No")

    def test_get_check_state(self):
        value = BoolValue()
        check_state = value.get_check_state(self.model, [0], [0])
        self.assertEqual(check_state, CheckState.CHECKED)

    def test_get_check_state_false(self):
        value = BoolValue()
        self.model.get_value = Mock(return_value=False)
        check_state = value.get_check_state(self.model, [0], [0])
        self.assertEqual(check_state, CheckState.UNCHECKED)

    def test_set_check_state(self):
        value = BoolValue()
        value.set_check_state(self.model, [0], [0], CheckState.CHECKED)
        self.model.set_value.assert_called_once_with([0], [0], True)

    def test_set_check_state_unchecked(self):
        value = BoolValue()
        value.set_check_state(self.model, [0], [0], CheckState.UNCHECKED)
        self.model.set_value.assert_called_once_with([0], [0], False)

    def test_set_check_state_no_set_value(self):
        self.model.can_set_value = Mock(return_value=False)
        value = BoolValue()
        with self.assertRaises(DataViewSetError):
            value.set_text(self.model, [0], [0], CheckState.CHECKED)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Enthought pyface package component
"""


import wx


from traits.api import Event, Property, Tuple, Str, VetoableEvent, provides


from pyface.i_window import IWindow, MWindow
from pyface.key_pressed_event import KeyPressedEvent
from .system_metrics import SystemMetrics
from .widget import Widget


@provides(IWindow)
class Window(MWindow, Widget):
    """ The toolkit specific implementation of a Window.  See the IWindow
    interface for the API documentation.
    """

    # 'IWindow' interface -----------------------------------------------------

    position = Property(Tuple)

    size = Property(Tuple)

    title = Str()

    # Window Events ----------------------------------------------------------

    #: The window has been opened.
    opened = Event()

    #: The window is about to open.
    opening = VetoableEvent()

    #: The window has been activated.
    activated = Event()

    #: The window has been closed.
    closed = Event()

    #: The window is about to be closed.
    closing = VetoableEvent()

    #: The window has been deactivated.
    deactivated = Event()

    #: A key was pressed while the window had focus.
    # FIXME v3: This smells of a hack. What's so special about key presses?
    # FIXME v3: Str
    key_pressed = Event(KeyPressedEvent)

    size = Property(Tuple)

    # Private interface ------------------------------------------------------

    # Shadow trait for position.
    _position = Tuple((-1, -1))

    # Shadow trait for size.
    _size = Tuple((-1, -1))

    # -------------------------------------------------------------------------
    # 'IWindow' interface.
    # -------------------------------------------------------------------------

    def activate(self):
        self.control.Iconize(False)
        self.control.Raise()

    def show(self, visible):
        self.control.Show(visible)

    # -------------------------------------------------------------------------
    # Protected 'IWindow' interface.
    # -------------------------------------------------------------------------

    def _add_event_listeners(self):
        self.control.Bind(wx.EVT_ACTIVATE, self._wx_on_activate)
        self.control.Bind(wx.EVT_SHOW, self._wx_on_show)
        self.control.Bind(wx.EVT_CLOSE, self._wx_on_close)
        self.control.Bind(wx.EVT_SIZE, self._wx_on_control_size)
        self.control.Bind(wx.EVT_MOVE, self._wx_on_control_move)
        self.control.Bind(wx.EVT_CHAR, self._wx_on_char)

    # -------------------------------------------------------------------------
    # Protected 'IWidget' interface.
    # -------------------------------------------------------------------------

    def _create_control(self, parent):
        # create a basic window control

        style = (
            wx.DEFAULT_FRAME_STYLE | wx.FRAME_NO_WINDOW_MENU | wx.CLIP_CHILDREN
        )
        control = wx.Frame(
            parent,
            -1,
            self.title,
            style=style,
            size=self.size,
            pos=self.position,
        )
        control.SetBackgroundColour(SystemMetrics().dialog_background_color)
        control.Enable(self.enabled)

        # XXX starting with self.visible true is generally a bad idea
        control.Show(self.visible)

        return control

    # -------------------------------------------------------------------------
    # Private interface.
    # -------------------------------------------------------------------------

    def _get_position(self):
        """ Property getter for position. """

        return self._position

    def _set_position(self, position):
        """ Property setter for position. """

        if self.control is not None:
            self.control.SetPosition(position)

        old = self._position
        self._position = position

        self.trait_property_changed("position", old, position)

    def _get_size(self):
        """ Property getter for size. """

        return self._size

    def _set_size(self, size):
        """ Property setter for size. """

        if self.control is not None:
            self.control.SetSize(size)

        old = self._size
        self._size = size

        self.trait_property_changed("size", old, size)

    def _title_changed(self, title):
        """ Static trait change handler. """

        if self.control is not None:
            self.control.SetTitle(title)

    # wx event handlers ------------------------------------------------------

    def _wx_on_activate(self, event):
        """ Called when the frame is being activated or deactivated. """

        if event.GetActive():
            self.activated = self
        else:
            self.deactivated = self

        event.Skip()

    def _wx_on_show(self, event):
        """ Called when the frame is being activated or deactivated. """

        self.visible = event.IsShown()

        event.Skip()

    def _wx_on_close(self, event):
        """ Called when the frame is being closed. """

        self.close()

    def _wx_on_control_move(self, event):
        """ Called when the window is resized. """

        # Get the real position and set the trait without performing
        # notification.

        # WXBUG - From the API documentation you would think that you could
        # call event.GetPosition directly, but that would be wrong.  The pixel
        # reported by that call is the pixel just below the window menu and
        # just right of the Windows-drawn border.

        try:
            self._position = (
                event.GetEventObject().GetPosition().Get()
            )  # Sizer.GetPosition().Get()
        except:
            pass
        event.Skip()

    def _wx_on_control_size(self, event):
        """ Called when the window is resized. """

        # Get the new size and set the shadow trait without performing
        # notification.
        wxsize = event.GetSize()

        self._size = (wxsize.GetWidth(), wxsize.GetHeight())

        event.Skip()

    def _wx_on_char(self, event):
        """ Called when a key is pressed when the tree has focus. """

        self.key_pressed = KeyPressedEvent(
            alt_down=event.altDown,
            control_down=event.controlDown,
            shift_down=event.shiftDown,
            key_code=event.KeyCode,
            event=event,
        )

        event.Skip()



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
# (C) Copyright 2007 Riverbank Computing Limited
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms described in the PyQt GPL exception also apply


from pyface.qt import QtGui, is_qt5


from traits.api import HasTraits, Int, Property, provides, Tuple


from pyface.i_system_metrics import ISystemMetrics, MSystemMetrics


@provides(ISystemMetrics)
class SystemMetrics(MSystemMetrics, HasTraits):
    """ The toolkit specific implementation of a SystemMetrics.  See the
    ISystemMetrics interface for the API documentation.
    """

    # 'ISystemMetrics' interface -------------------------------------------

    screen_width = Property(Int)

    screen_height = Property(Int)

    dialog_background_color = Property(Tuple)

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _get_screen_width(self):
        # QDesktopWidget.screenGeometry() is deprecated and Qt docs
        # suggest using screens() instead, but screens in not available in qt4
        # see issue: enthought/pyface#721
        if is_qt5:
            return QtGui.QApplication.instance().screens()[0].availableGeometry().width()
        else:
            return QtGui.QApplication.instance().desktop().availableGeometry().width()

    def _get_screen_height(self):
        # QDesktopWidget.screenGeometry(int screen) is deprecated and Qt docs
        # suggest using screens() instead, but screens in not available in qt4
        # see issue: enthought/pyface#721
        if is_qt5:
            return (
                QtGui.QApplication.instance().screens()[0].availableGeometry().height()
            )
        else:
            return (
                QtGui.QApplication.instance().desktop().availableGeometry().height()
            )

    def _get_dialog_background_color(self):
        color = (
            QtGui.QApplication.instance()
            .palette()
            .color(QtGui.QPalette.Window)
        )

        return (color.redF(), color.greenF(), color.blueF())



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" A renderer which displays a checked-box for a True value and an unchecked
    box for a false value. """


from pyface.image_resource import ImageResource


from .mapped_grid_cell_image_renderer import MappedGridCellImageRenderer

checked_image_map = {
    True: ImageResource("checked"),
    False: ImageResource("unchecked"),
}


class CheckboxImageRenderer(MappedGridCellImageRenderer):
    def __init__(self, display_text=False):

        text_map = None
        if display_text:
            text_map = {True: "True", False: "False"}

        # Base-class constructor
        super(CheckboxImageRenderer, self).__init__(
            checked_image_map, text_map
        )

        return



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


import unittest

from ..system_metrics import SystemMetrics


class TestSystemMetrics(unittest.TestCase):
    def setUp(self):
        self.metrics = SystemMetrics()

    def test_width(self):
        width = self.metrics.screen_width
        self.assertGreaterEqual(width, 0)

    def test_height(self):
        height = self.metrics.screen_height
        self.assertGreaterEqual(height, 0)

    def test_background_color(self):
        color = self.metrics.dialog_background_color
        self.assertIn(len(color), [3, 4])
        self.assertTrue(all(0 <= channel <= 1 for channel in color))



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
# (C) Copyright 2007 Riverbank Computing Limited
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms described in the PyQt GPL exception also apply


from pyface.qt import QtCore, QtGui


from traits.api import (
    Any, Bool, Callable, Enum, Int, List, provides, Str, Tuple
)

from pyface.i_dialog import IDialog, MDialog
from pyface.constant import OK, CANCEL, YES, NO
from .window import Window


# Map PyQt dialog related constants to the pyface equivalents.
_RESULT_MAP = {
    int(QtGui.QDialog.Accepted): OK,
    int(QtGui.QDialog.Rejected): CANCEL,
    int(QtGui.QMessageBox.Ok): OK,
    int(QtGui.QMessageBox.Cancel): CANCEL,
    int(QtGui.QMessageBox.Yes): YES,
    int(QtGui.QMessageBox.No): NO,
}


@provides(IDialog)
class Dialog(MDialog, Window):
    """ The toolkit specific implementation of a Dialog.  See the IDialog
    interface for the API documentation.
    """

    # 'IDialog' interface -------------------------------------------------#

    cancel_label = Str()

    help_id = Str()

    help_label = Str()

    ok_label = Str()

    resizeable = Bool(True)

    return_code = Int(OK)

    style = Enum("modal", "nonmodal")

    # 'IWindow' interface -------------------------------------------------#

    title = Str("Dialog")

    # Private interface ---------------------------------------------------#

    #: A list of connected Qt signals to be removed before destruction.
    #: First item in the tuple is the Qt signal. The second item is the event
    #: handler.
    _connections_to_remove = List(Tuple(Any, Callable))

    # ------------------------------------------------------------------------
    # Protected 'IDialog' interface.
    # ------------------------------------------------------------------------

    def _create_buttons(self, parent):
        buttons = QtGui.QDialogButtonBox()

        # 'OK' button.
        if self.ok_label:
            btn = buttons.addButton(
                self.ok_label, QtGui.QDialogButtonBox.AcceptRole
            )
        else:
            btn = buttons.addButton(QtGui.QDialogButtonBox.Ok)

        btn.setDefault(True)
        btn.clicked.connect(self.control.accept)
        self._connections_to_remove.append((btn.clicked, self.control.accept))

        # 'Cancel' button.
        if self.cancel_label:
            btn = buttons.addButton(
                self.cancel_label, QtGui.QDialogButtonBox.RejectRole
            )
        else:
            btn = buttons.addButton(QtGui.QDialogButtonBox.Cancel)

        btn.clicked.connect(self.control.reject)
        self._connections_to_remove.append((btn.clicked, self.control.reject))

        # 'Help' button.
        # FIXME v3: In the original code the only possible hook into the help
        # was to reimplement self._on_help().  However this was a private
        # method.  Obviously nobody uses the Help button.  For the moment we
        # display it but can't actually use it.
        if len(self.help_id) > 0:
            if self.help_label:
                buttons.addButton(
                    self.help_label, QtGui.QDialogButtonBox.HelpRole
                )
            else:
                buttons.addButton(QtGui.QDialogButtonBox.Help)

        return buttons

    def _create_contents(self, parent):
        layout = QtGui.QVBoxLayout()

        if not self.resizeable:
            layout.setSizeConstraint(QtGui.QLayout.SetFixedSize)

        layout.addWidget(self._create_dialog_area(parent))
        layout.addWidget(self._create_buttons(parent))

        parent.setLayout(layout)

    def _create_dialog_area(self, parent):
        panel = QtGui.QWidget(parent)
        panel.setMinimumSize(QtCore.QSize(100, 200))

        palette = panel.palette()
        palette.setColor(QtGui.QPalette.Window, QtGui.QColor("red"))
        panel.setPalette(palette)
        panel.setAutoFillBackground(True)

        return panel

    def _show_modal(self):
        dialog = self.control
        dialog.setWindowModality(QtCore.Qt.ApplicationModal)

        # Suppress the context-help button hint, which
        # results in a non-functional "?" button on Windows.
        dialog.setWindowFlags(
            dialog.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint
        )

        retval = dialog.exec_()
        return _RESULT_MAP[retval]

    # -------------------------------------------------------------------------
    # 'IWidget' interface.
    # -------------------------------------------------------------------------

    def destroy(self):
        while self._connections_to_remove:
            signal, handler = self._connections_to_remove.pop()
            signal.disconnect(handler)

        super(Dialog, self).destroy()

    # ------------------------------------------------------------------------
    # Protected 'IWidget' interface.
    # ------------------------------------------------------------------------

    def _create_control(self, parent):
        dlg = QtGui.QDialog(parent)

        # Setting return code and firing close events is handled for 'modal' in
        # MDialog's open method. For 'nonmodal', we do it here.
        if self.style == "nonmodal":
            dlg.finished.connect(self._finished_fired)
            self._connections_to_remove.append(
                (dlg.finished, self._finished_fired)
            )

        if self.size != (-1, -1):
            dlg.resize(*self.size)

        if self.position != (-1, -1):
            dlg.move(*self.position)

        dlg.setWindowTitle(self.title)

        return dlg

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _finished_fired(self, result):
        """ Called when the dialog is closed (and nonmodal). """

        self.return_code = _RESULT_MAP[result]
        self.close()



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" The perspective interface. """


from traits.api import Bool, Interface, List, Str, Tuple


from .perspective_item import PerspectiveItem


class IPerspective(Interface):
    """ The perspective interface. """

    # The perspective's unique identifier (unique within a workbench window).
    id = Str()

    # The perspective's name.
    name = Str()

    # The contents of the perspective.
    contents = List(PerspectiveItem)

    # The size of the editor area in this perspective. A value of (-1, -1)
    # indicates that the workbench window should choose an appropriate size
    # based on the sizes of the views in the perspective.
    editor_area_size = Tuple()

    # Is the perspective enabled?
    enabled = Bool()

    # Should the editor area be shown in this perspective?
    show_editor_area = Bool()

    # Methods -------------------------------------------------------------#

    def create(self, window):
        """ Create the perspective in a workbench window. """

    def show(self, window):
        """ Called when the perspective is shown in a workbench window. """



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The 'null' backend specific implementation of a menu manager.
"""


from traits.api import Str


from pyface.action.action_manager import ActionManager
from pyface.action.action_manager_item import ActionManagerItem
from pyface.action.group import Group


class MenuManager(ActionManager, ActionManagerItem):
    """ A menu manager realizes itself in a menu control.

    This could be a sub-menu or a context (popup) menu.
    """

    # 'MenuManager' interface ---------------------------------------------#

    # The menu manager's name (if the manager is a sub-menu, this is what its
    # label will be).
    name = Str()

    # ------------------------------------------------------------------------
    # 'MenuManager' interface.
    # ------------------------------------------------------------------------

    def create_menu(self, parent, controller=None):
        """ Creates a menu representation of the manager. """

        # If a controller is required it can either be set as a trait on the
        # menu manager (the trait is part of the 'ActionManager' API), or
        # passed in here (if one is passed in here it takes precedence over the
        # trait).
        if controller is None:
            controller = self.controller

        return None

    # ------------------------------------------------------------------------
    # 'ActionManagerItem' interface.
    # ------------------------------------------------------------------------

    def add_to_menu(self, parent, menu, controller):
        """ Adds the item to a menu. """

    def add_to_toolbar(self, parent, tool_bar, image_cache, controller):
        """ Adds the item to a tool bar. """

        raise ValueError("Cannot add a menu manager to a toolbar.")



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" A generic base-class for items in a tree data structure.

An example:-

root = TreeItem(data='Root')

fruit = TreeItem(data='Fruit')
fruit.append(TreeItem(data='Apple', allows_children=False))
fruit.append(TreeItem(data='Orange', allows_children=False))
fruit.append(TreeItem(data='Pear', allows_children=False))
root.append(fruit)

veg = TreeItem(data='Veg')
veg.append(TreeItem(data='Carrot', allows_children=False))
veg.append(TreeItem(data='Cauliflower', allows_children=False))
veg.append(TreeItem(data='Sprout', allows_children=False))
root.append(veg)

"""


from traits.api import Any, Bool, HasTraits, Instance, List, Property


class TreeItem(HasTraits):
    """ A generic base-class for items in a tree data structure. """

    # 'TreeItem' interface -------------------------------------------------

    # Does this item allow children?
    allows_children = Bool(True)

    # The item's children.
    children = List(Instance("TreeItem"))

    # Arbitrary data associated with the item.
    data = Any()

    # Does the item have any children?
    has_children = Property(Bool)

    # The item's parent.
    parent = Instance("TreeItem")

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __str__(self):
        """ Returns the informal string representation of the object. """

        if self.data is None:
            s = ""

        else:
            s = str(self.data)

        return s

    # ------------------------------------------------------------------------
    # 'TreeItem' interface.
    # ------------------------------------------------------------------------

    # Properties -----------------------------------------------------------

    # has_children
    def _get_has_children(self):
        """ True iff the item has children. """

        return len(self.children) != 0

    # Methods -------------------------------------------------------------#

    def append(self, child):
        """ Appends a child to this item.

        This removes the child from its current parent (if it has one).

        """

        return self.insert(len(self.children), child)

    def insert(self, index, child):
        """ Inserts a child into this item at the specified index.

        This removes the child from its current parent (if it has one).

        """

        if child.parent is not None:
            child.parent.remove(child)

        child.parent = self
        self.children.insert(index, child)

        return child

    def remove(self, child):
        """ Removes a child from this item. """

        child.parent = None
        self.children.remove(child)

        return child

    def insert_before(self, before, child):
        """ Inserts a child into this item before the specified item.

        This removes the child from its current parent (if it has one).

        """

        index = self.children.index(before)

        self.insert(index, child)

        return (index, child)

    def insert_after(self, after, child):
        """ Inserts a child into this item after the specified item.

        This removes the child from its current parent (if it has one).

        """

        index = self.children.index(after)

        self.insert(index + 1, child)

        return (index, child)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The interface for workbench parts. """


from traits.api import Any, Bool, HasTraits, Instance, Interface
from traits.api import List, provides, Str


class IWorkbenchPart(Interface):
    """ The interface for workbench parts.

    A workbench part is a visual section within the workbench. There are two
    sub-types, 'View' and 'Editor'.

    """

    # The toolkit-specific control that represents the part.
    #
    # The framework sets this to the value returned by 'create_control'.
    control = Any()

    # Does the part currently have the focus?
    has_focus = Bool(False)

    # The part's globally unique identifier.
    id = Str()

    # The part's name (displayed to the user).
    name = Str()

    # The current selection within the part.
    selection = List()

    # The workbench window that the part is in.
    #
    # The framework sets this when the part is created.
    window = Instance("pyface.workbench.api.WorkbenchWindow")

    # Methods -------------------------------------------------------------#

    def create_control(self, parent):
        """ Create the toolkit-specific control that represents the part.

        The parameter *parent* is the toolkit-specific control that is the
        parts's parent.

        Return the toolkit-specific control.

        """

    def destroy_control(self):
        """ Destroy the toolkit-specific control that represents the part.

        Return None.

        """

    def set_focus(self):
        """ Set the focus to the appropriate control in the part.

        Return None.

        """


@provides(IWorkbenchPart)
class MWorkbenchPart(HasTraits):
    """ Mixin containing common code for toolkit-specific implementations. """

    # 'IWorkbenchPart' interface -------------------------------------------

    # The toolkit-specific control that represents the part.
    #
    # The framework sets this to the value returned by 'create_control'.
    control = Any()

    # Does the part currently have the focus?
    has_focus = Bool(False)

    # The part's globally unique identifier.
    id = Str()

    # The part's name (displayed to the user).
    name = Str()

    # The current selection within the part.
    selection = List()

    # The workbench window that the part is in.
    #
    # The framework sets this when the part is created.
    window = Instance("pyface.workbench.api.WorkbenchWindow")

    # Methods -------------------------------------------------------------#

    def create_control(self, parent):
        """ Create the toolkit-specific control that represents the part. """

        raise NotImplementedError()

    def destroy_control(self):
        """ Destroy the toolkit-specific control that represents the part. """

        raise NotImplementedError()

    def set_focus(self):
        """ Set the focus to the appropriate control in the part. """

        raise NotImplementedError()



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Routines supporting color computations

Most of what is needed is provided by Python's builtin colorsys module,
but we need a few additional routines for things that are not covered by
that code.
"""


def channels_to_ints(channels, maximum=255):
    """ Convert an iterable of floating point channel values to integers.

    Values are rounded to the nearest integer, rather than truncated.

    Parameters
    ----------
    channels : iterable of float
        An iterable of channel values, each value between 0.0 and 1.0,
        inclusive.
    maximum : int
        The maximum value of the integer range.  Common values are 15,
        65535 or 255, which is the default.

    Returns
    -------
    values : tuple of int
        A tuple of values as integers between 0 and max, inclusive.
    """
    return tuple(int(round(channel * maximum)) for channel in channels)


def ints_to_channels(values, maximum=255):
    """ Convert an iterable of integers to floating point channel values.

    Parameters
    ----------
    values : tuple of int
        An iterable of values as integers between 0 and max, inclusive.
    maximum : int
        The maximum value of the integer range.  Common values are 15,
        65535 or 255, which is the default.

    Returns
    -------
    channels : iterable of float
        A tuple of channel values, each value between 0.0 and 1.0,
        inclusive.
    """
    return tuple(value / maximum for value in values)


def relative_luminance(rgb):
    """ The relative luminance of the color.

    This value is the critical value when comparing colors for contrast when
    displayed next to each other, in particular for readability of text.

    Parameters
    ----------
    rgb : tuple of red, green, blue values
        A tuple of values representing red, green and blue components of
        the color, as values from 0.0 to 1.0.

    Returns
    -------
    luminance : float
        The relative luminance of the color.

    References
    ----------
    Web Contrast Accessibility Guidelines
    https://www.w3.org/TR/2008/REC-WCAG20-20081211/#relativeluminancedef
    """
    gamma_corrected = [
        x/12.92 if x <= 0.03928 else ((x + 0.055)/1.055)**2.4
        for x in rgb
    ]
    luminance = (
        0.2126 * gamma_corrected[0]
        + 0.7152 * gamma_corrected[1]
        + 0.0722 * gamma_corrected[2]
    )
    return luminance


def is_dark(rgb):
    """ Is the color dark to human perception?

    A color is dark if white contasts better with it according to the WC3
    definition of contrast ratio.  This is allows GUI code to choose either
    black or white as a contrasting color for things like text on a colored
    background.

    Parameters
    ----------
    rgb : tuple of red, green, blue values
        A tuple of values representing red, green and blue components of
        the color, as values from 0.0 to 1.0.

    References
    ----------
    Understanding Web Contrast Accessibility Guidelines
    https://www.w3.org/TR/UNDERSTANDING-WCAG20/visual-audio-contrast-contrast.html#contrast-ratiodef
    """
    lumininance = relative_luminance(rgb)
    black_contrast = (lumininance + 0.05) / 0.05
    white_contrast = 1.05 / (lumininance + 0.05)
    return white_contrast > black_contrast



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


import weakref

from pyface.qt import QtGui, QtCore

from .find_widget import FindWidget


class ReplaceWidget(FindWidget):
    def __init__(self, parent):
        super(FindWidget, self).__init__(parent)
        self.adv_code_widget = weakref.ref(parent)

        # QFontMetrics.width() is deprecated and Qt docs suggest using
        # horizontalAdvance() instead, but is only available since Qt 5.11
        if QtCore.__version_info__ >= (5, 11):
            self.button_size = self.fontMetrics().horizontalAdvance("Replace All") + 20
        else:
            self.button_size = self.fontMetrics().width("Replace All") + 20

        form_layout = QtGui.QFormLayout()
        form_layout.addRow("Fin&d", self._create_find_control())
        form_layout.addRow("Rep&lace", self._create_replace_control())

        layout = QtGui.QHBoxLayout()
        layout.addLayout(form_layout)

        close_button = QtGui.QPushButton("Close")
        layout.addWidget(close_button, 1, QtCore.Qt.AlignRight)
        close_button.clicked.connect(self.hide)

        self.setLayout(layout)

    def _create_replace_control(self):
        control = QtGui.QWidget(self)

        self.replace_edit = QtGui.QLineEdit()
        self.replace_button = QtGui.QPushButton("&Replace")
        self.replace_button.setFixedWidth(self.button_size)
        self.replace_all_button = QtGui.QPushButton("Replace &All")
        self.replace_all_button.setFixedWidth(self.button_size)

        layout = QtGui.QHBoxLayout()
        layout.addWidget(self.replace_edit)
        layout.addWidget(self.replace_button)
        layout.addWidget(self.replace_all_button)
        layout.addStretch(2)
        layout.setContentsMargins(0, 0, 0, 0)

        control.setLayout(layout)
        return control



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


import math

from pyface.qt import QtCore, QtGui


class GutterWidget(QtGui.QWidget):

    min_width = 5
    background_color = QtGui.QColor(220, 220, 220)

    def sizeHint(self):
        return QtCore.QSize(self.min_width, 0)

    def paintEvent(self, event):
        """ Paint the line numbers.
        """
        painter = QtGui.QPainter(self)
        painter.fillRect(event.rect(), QtCore.Qt.lightGray)

    def wheelEvent(self, event):
        """ Delegate mouse wheel events to parent for seamless scrolling.
        """
        self.parent().wheelEvent(event)


class StatusGutterWidget(GutterWidget):
    """ Draws status markers
    """

    def __init__(self, *args, **kw):
        super(StatusGutterWidget, self).__init__(*args, **kw)

        self.error_lines = []
        self.warn_lines = []
        self.info_lines = []

    def sizeHint(self):
        return QtCore.QSize(10, 0)

    def paintEvent(self, event):
        """ Paint the line numbers.
        """
        painter = QtGui.QPainter(self)
        painter.fillRect(event.rect(), self.background_color)

        cw = self.parent()

        pixels_per_block = self.height() / float(cw.blockCount())

        for line in self.info_lines:
            painter.fillRect(
                QtCore.QRect(0, line * pixels_per_block, self.width(), 3),
                QtCore.Qt.green,
            )

        for line in self.warn_lines:
            painter.fillRect(
                QtCore.QRect(0, line * pixels_per_block, self.width(), 3),
                QtCore.Qt.yellow,
            )

        for line in self.error_lines:
            painter.fillRect(
                QtCore.QRect(0, line * pixels_per_block, self.width(), 3),
                QtCore.Qt.red,
            )


class LineNumberWidget(GutterWidget):
    """ Draw line numbers.
    """

    min_char_width = 4

    def fontMetrics(self):
        # QWidget's fontMetrics method does not provide an up to date
        # font metrics, just one corresponding to the initial font
        return QtGui.QFontMetrics(self.font)

    def set_font(self, font):
        self.font = font

    def digits_width(self):
        nlines = max(1, self.parent().blockCount())
        ndigits = max(
            self.min_char_width, int(math.floor(math.log10(nlines) + 1))
        )
        # QFontMetrics.width() is deprecated and Qt docs suggest using
        # horizontalAdvance() instead, but is only available since Qt 5.11
        if QtCore.__version_info__ >= (5, 11):
            width = max(
                self.fontMetrics().horizontalAdvance("0" * ndigits) + 3,
                self.min_width
            )
        else:
            width = max(
                self.fontMetrics().width("0" * ndigits) + 3, self.min_width
            )
        return width

    def sizeHint(self):
        return QtCore.QSize(self.digits_width(), 0)

    def paintEvent(self, event):
        """ Paint the line numbers.
        """
        painter = QtGui.QPainter(self)
        painter.setFont(self.font)
        painter.fillRect(event.rect(), self.background_color)

        cw = self.parent()
        block = cw.firstVisibleBlock()
        blocknum = block.blockNumber()
        top = (
            cw.blockBoundingGeometry(block)
            .translated(cw.contentOffset())
            .top()
        )
        bottom = top + int(cw.blockBoundingRect(block).height())

        while block.isValid() and top <= event.rect().bottom():
            if block.isVisible() and bottom >= event.rect().top():
                painter.setPen(QtCore.Qt.black)
                painter.drawText(
                    0,
                    top,
                    self.width() - 2,
                    self.fontMetrics().height(),
                    QtCore.Qt.AlignRight,
                    str(blocknum + 1),
                )
            block = block.next()
            top = bottom
            bottom = top + int(cw.blockBoundingRect(block).height())
            blocknum += 1
